from pony.py23compat import PY2

import unittest
from pony.orm.tests.testutils import *
from pony.orm.tests import db_params, setup_database, teardown_database

from pony.orm import *

db = Database()


class Foo(db.Entity):
    id = PrimaryKey(int)
    a = Required(int)
    b = Required(int)
    c = Required(int)
    array1 = Required(IntArray, index=True)
    array2 = Required(FloatArray)
    array3 = Required(StrArray)
    array4 = Optional(IntArray)
    array5 = Optional(IntArray, nullable=True)


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if db_params['provider'] not in ('sqlite', 'postgres'):
            raise unittest.SkipTest('Arrays are only available for SQLite and PostgreSQL')

        setup_database(db)
        with db_session:
            Foo(id=1, a=1, b=3, c=-2, array1=[10, 20, 30, 40, 50], array2=[1.1, 2.2, 3.3, 4.4, 5.5],
                array3=['foo', 'bar'])

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    @db_session
    def test_1(self):
        foo = select(f for f in Foo if 10 in f.array1)[:]
        self.assertEqual([Foo[1]], foo)

    @db_session
    def test_2(self):
        foo = select(f for f in Foo if [10, 20, 50] in f.array1)[:]
        self.assertEqual([Foo[1]], foo)

    @db_session
    def test_2a(self):
        foo = select(f for f in Foo if [] in f.array1)[:]
        self.assertEqual([Foo[1]], foo)

    @db_session
    def test_3(self):
        x = [10, 20, 50]
        foo = select(f for f in Foo if x in f.array1)[:]
        self.assertEqual([Foo[1]], foo)

    @db_session
    def test_4(self):
        foo = select(f for f in Foo if 1.1 in f.array2)[:]
        self.assertEqual([Foo[1]], foo)

    err_msg = "Cannot store 'int' item in array of " + ("'unicode'" if PY2 else "'str'")

    @raises_exception(TypeError, err_msg)
    @db_session
    def test_5(self):
        foo = Foo.select().first()
        foo.array3.append(123)

    @raises_exception(TypeError, err_msg)
    @db_session
    def test_6(self):
        foo = Foo.select().first()
        foo.array3[0] = 123

    @raises_exception(TypeError, err_msg)
    @db_session
    def test_7(self):
        foo = Foo.select().first()
        foo.array3.extend(['str', 123, 'str'])

    @db_session
    def test_8(self):
        foo = Foo.select().first()
        foo.array3.extend(['str1', 'str2'])

    @db_session
    def test_9(self):
        foos = select(f.array2[0] for f in Foo)[:]
        self.assertEqual([1.1], foos)

    @db_session
    def test_10(self):
        foos = select(f.array1[1:-1] for f in Foo)[:]
        self.assertEqual([20, 30, 40], foos[0])

    @db_session
    def test_11(self):
        foo = Foo.select().first()
        foo.array4.append(1)
        self.assertEqual([1], foo.array4)

    @raises_exception(AttributeError, "'NoneType' object has no attribute 'append'")
    @db_session
    def test_12(self):
        foo = Foo.select().first()
        foo.array5.append(1)

    @db_session
    def test_13(self):
        x = [10, 20, 30, 40, 50]
        ids = select(f.id for f in Foo if x == f.array1)[:]
        self.assertEqual(ids, [1])

    @db_session
    def test_14(self):
        val = select(f.array1[0] for f in Foo).first()
        self.assertEqual(val, 10)

    @db_session
    def test_15(self):
        val = select(f.array1[2] for f in Foo).first()
        self.assertEqual(val, 30)

    @db_session
    def test_16(self):
        val = select(f.array1[-1] for f in Foo).first()
        self.assertEqual(val, 50)

    @db_session
    def test_17(self):
        val = select(f.array1[-2] for f in Foo).first()
        self.assertEqual(val, 40)

    @db_session
    def test_18(self):
        x = 2
        val = select(f.array1[x] for f in Foo).first()
        self.assertEqual(val, 30)

    @db_session
    def test_19(self):
        val = select(f.array1[f.a] for f in Foo).first()
        self.assertEqual(val, 20)

    @db_session
    def test_20(self):
        val = select(f.array1[f.c] for f in Foo).first()
        self.assertEqual(val, 40)

    @db_session
    def test_21(self):
        array = select(f.array1[2:4] for f in Foo).first()
        self.assertEqual(array, [30, 40])

    @db_session
    def test_22(self):
        array = select(f.array1[1:-2] for f in Foo).first()
        self.assertEqual(array, [20, 30])

    @db_session
    def test_23(self):
        array = select(f.array1[10:-10] for f in Foo).first()
        self.assertEqual(array, [])

    @db_session
    def test_24(self):
        x = 2
        array = select(f.array1[x:4] for f in Foo).first()
        self.assertEqual(array, [30, 40])

    @db_session
    def test_25(self):
        y = 4
        array = select(f.array1[2:y] for f in Foo).first()
        self.assertEqual(array, [30, 40])

    @db_session
    def test_26(self):
        x, y = 2, 4
        array = select(f.array1[x:y] for f in Foo).first()
        self.assertEqual(array, [30, 40])

    @db_session
    def test_27(self):
        x, y = 1, -2
        array = select(f.array1[x:y] for f in Foo).first()
        self.assertEqual(array, [20, 30])

    @db_session
    def test_28(self):
        x = 1
        array = select(f.array1[x:f.b] for f in Foo).first()
        self.assertEqual(array, [20, 30])

    @db_session
    def test_29(self):
        array = select(f.array1[f.a:f.c] for f in Foo).first()
        self.assertEqual(array, [20, 30])

    @db_session
    def test_30(self):
        array = select(f.array1[:3] for f in Foo).first()
        self.assertEqual(array, [10, 20, 30])

    @db_session
    def test_31(self):
        array = select(f.array1[2:] for f in Foo).first()
        self.assertEqual(array, [30, 40, 50])

    @db_session
    def test_32(self):
        array = select(f.array1[:f.b] for f in Foo).first()
        self.assertEqual(array, [10, 20, 30])

    @db_session
    def test_33(self):
        array = select(f.array1[:f.c] for f in Foo).first()
        self.assertEqual(array, [10, 20, 30])

    @db_session
    def test_34(self):
        array = select(f.array1[f.c:] for f in Foo).first()
        self.assertEqual(array, [40, 50])

    @db_session
    def test_35(self):
        foo = Foo.select().first()
        self.assertTrue(10 in foo.array1)
        self.assertTrue(1000 not in foo.array1)
        self.assertTrue([10, 20] in foo.array1)
        self.assertTrue([20, 10] in foo.array1)
        self.assertTrue([10, 1000] not in foo.array1)
        self.assertTrue([] in foo.array1)
        self.assertTrue('bar' in foo.array3)
        self.assertTrue('baz' not in foo.array3)
        self.assertTrue(['foo', 'bar'] in foo.array3)
        self.assertTrue(['bar', 'foo'] in foo.array3)
        self.assertTrue(['baz', 'bar'] not in foo.array3)
        self.assertTrue([] in foo.array3)

    @db_session
    def test_36(self):
        items = []
        result = select(foo for foo in Foo if foo in items)[:]
        self.assertEqual(result, [])

    @db_session
    def test_37(self):
        f1 = Foo[1]
        items = [f1]
        result = select(foo for foo in Foo if foo in items)[:]
        self.assertEqual(result, [f1])

    @db_session
    def test_38(self):
        items = []
        result = select(foo for foo in Foo if foo.id in items)[:]
        self.assertEqual(result, [])

    @db_session
    def test_39(self):
        items = [1]
        result = select(foo.id for foo in Foo if foo.id in items)[:]
        self.assertEqual(result, [1])



from __future__ import absolute_import, print_function, division
from pony.py23compat import PYPY, PYPY2

import sys, unittest
from datetime import date
from decimal import Decimal

from pony.orm.core import *
from pony.orm.sqltranslation import IncomparableTypesError
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Student(db.Entity):
    name = Required(unicode)
    dob = Optional(date)
    gpa = Optional(float)
    scholarship = Optional(Decimal, 7, 2)
    group = Required('Group')
    courses = Set('Course')

class Group(db.Entity):
    number = PrimaryKey(int)
    students = Set(Student)
    dept = Required('Department')

class Department(db.Entity):
    number = PrimaryKey(int)
    groups = Set(Group)

class Course(db.Entity):
    name = Required(unicode)
    semester = Required(int)
    PrimaryKey(name, semester)
    students = Set(Student)


class TestSQLTranslatorExceptions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            d1 = Department(number=44)
            g1 = Group(number=101, dept=d1)
            Student(name='S1', group=g1)
            Student(name='S2', group=g1)
            Student(name='S3', group=g1)
    @classmethod
    def tearDownClass(cls):
        teardown_database(db)
    def setUp(self):
        rollback()
        db_session.__enter__()
    def tearDown(self):
        rollback()
        db_session.__exit__()
    @raises_exception(NotImplementedError, 'for x in s.name')
    def test1(self):
        x = 10
        select(s for s in Student for x in s.name)
    @raises_exception(TranslationError, "Inside declarative query, iterator must be entity or query. Got: for i in x")
    def test2(self):
        x = [1, 2, 3]
        select(s for s in Student for i in x)
    @raises_exception(TranslationError, "Inside declarative query, iterator must be entity or query. Got: for s2 in g.students")
    def test3(self):
        g = Group[101]
        select(s for s in Student for s2 in g.students)
    @raises_exception(NotImplementedError, "*args is not supported")
    def test4(self):
        args = 'abc'
        select(s for s in Student if s.name.upper(*args))

    if sys.version_info[:2] < (3, 5): # TODO
        @raises_exception(NotImplementedError) # "**{'a': 'b', 'c': 'd'} is not supported
        def test5(self):
            select(s for s in Student if s.name.upper(**{'a':'b', 'c':'d'}))

    @raises_exception(ExprEvalError, "`1 in 2` raises TypeError: argument of type 'int' is not iterable" if not PYPY else
                                     "`1 in 2` raises TypeError: 'int' object is not iterable")
    def test6(self):
        select(s for s in Student if 1 in 2)
    @raises_exception(NotImplementedError, 'Group[s.group.number]')
    def test7(self):
        select(s for s in Student if Group[s.group.number].dept.number == 44)
    @raises_exception(ExprEvalError, "`Group[123, 456].dept.number == 44` raises TypeError: Invalid count of attrs in Group primary key (2 instead of 1)")
    def test8(self):
        select(s for s in Student if Group[123, 456].dept.number == 44)
    @raises_exception(ExprEvalError, "`Course[123]` raises TypeError: Invalid count of attrs in Course primary key (1 instead of 2)")
    def test9(self):
        select(s for s in Student if Course[123] in s.courses)
    @raises_exception(TypeError, "Incomparable types '%s' and 'float' in expression: s.name < s.gpa" % unicode.__name__)
    def test10(self):
        select(s for s in Student if s.name < s.gpa)
    @raises_exception(ExprEvalError, "`Group(101)` raises TypeError: Group constructor accept only keyword arguments. Got: 1 positional argument")
    def test11(self):
        select(s for s in Student if s.group == Group(101))
    @raises_exception(ExprEvalError, "`Group[date(2011, 1, 2)]` raises TypeError: Value type for attribute Group.number must be int. Got: %r" % date)
    def test12(self):
        select(s for s in Student if s.group == Group[date(2011, 1, 2)])
    @raises_exception(TypeError, "Unsupported operand types 'int' and '%s' for operation '+' in expression: s.group.number + s.name" % unicode.__name__)
    def test13(self):
        select(s for s in Student if s.group.number + s.name < 0)
    @raises_exception(TypeError, "Unsupported operand types 'Decimal' and 'float' for operation '+' in expression: s.scholarship + 1.1")
    def test14(self):
        select(s for s in Student if s.scholarship + 1.1 > 10)
    @raises_exception(TypeError, "Unsupported operand types 'Decimal' and '%s' for operation '**' "
                                 "in expression: s.scholarship ** 'abc'" % unicode.__name__)
    def test15(self):
        select(s for s in Student if s.scholarship ** 'abc' > 10)
    @raises_exception(TypeError, "Unsupported operand types '%s' and 'int' for operation '+' in expression: s.name + 2" % unicode.__name__)
    def test16(self):
        select(s for s in Student if s.name + 2 > 10)
    @raises_exception(TypeError, "Step is not supported in s.name[1:3:5]")
    def test17(self):
        select(s for s in Student if s.name[1:3:5] == 'A')
    @raises_exception(TypeError, "Invalid type of start index (expected 'int', got '%s') in string slice s.name['a':1]"
                                 % unicode.__name__)
    def test18(self):
        select(s for s in Student if s.name['a':1] == 'A')
    @raises_exception(TypeError, "Invalid type of stop index (expected 'int', got '%s') in string slice s.name[1:'a']"
                                 % unicode.__name__)
    def test19(self):
        select(s for s in Student if s.name[1:'a'] == 'A')
    @raises_exception(TypeError, "String indices must be integers. Got '%s' in expression s.name['a']" % unicode.__name__)
    def test21(self):
        select(s.name for s in Student if s.name['a'] == 'h')
    @raises_exception(TypeError, "Incomparable types 'int' and '%s' in expression: 1 in s.name" % unicode.__name__)
    def test22(self):
        select(s.name for s in Student if 1 in s.name)
    @raises_exception(TypeError, "Expected '%s' argument but got 'int' in expression s.name.startswith(1)" % unicode.__name__)
    def test23(self):
        select(s.name for s in Student if s.name.startswith(1))
    @raises_exception(TypeError, "Expected '%s' argument but got 'int' in expression s.name.endswith(1)" % unicode.__name__)
    def test24(self):
        select(s.name for s in Student if s.name.endswith(1))
    @raises_exception(TypeError, "'chars' argument must be of '%s' type in s.name.strip(1), got: 'int'" % unicode.__name__)
    def test25(self):
        select(s.name for s in Student if s.name.strip(1))
    @raises_exception(AttributeError, "'%s' object has no attribute 'unknown': s.name.unknown" % unicode.__name__)
    def test26(self):
        result = set(select(s for s in Student if s.name.unknown() == "joe"))
    @raises_exception(AttributeError, "Entity Group does not have attribute foo: s.group.foo")
    def test27(self):
        select(s.name for s in Student if s.group.foo.bar == 10)
    @raises_exception(ExprEvalError, "`g.dept.foo.bar` raises AttributeError: 'Department' object has no attribute 'foo'")
    def test28(self):
        g = Group[101]
        select(s for s in Student if s.name == g.dept.foo.bar)
    @raises_exception(TypeError, "'year' argument of date(year, month, day) function must be of 'int' type. "
                                 "Got: '%s'" % unicode.__name__)
    def test29(self):
        select(s for s in Student if s.dob < date('2011', 1, 1))
    @raises_exception(NotImplementedError, "date(s.id, 1, 1)")
    def test30(self):
        select(s for s in Student if s.dob < date(s.id, 1, 1))
    @raises_exception(
        ExprEvalError,
        "`max()` raises TypeError: max() expects at least one argument" if PYPY else
        "`max()` raises TypeError: max expected 1 arguments, got 0" if sys.version_info[:2] < (3, 8) else
        "`max()` raises TypeError: max expected 1 argument, got 0" if sys.version_info[:2] < (3, 9) else
        "`max()` raises TypeError: max expected at least 1 argument, got 0")
    def test31(self):
        select(s for s in Student if s.id < max())
    @raises_exception(TypeError, "Incomparable types 'Student' and 'Course' in expression: s in s.courses")
    def test32(self):
        select(s for s in Student if s in s.courses)
    @raises_exception(AttributeError, "s.courses.name.foo")
    def test33(self):
        select(s for s in Student if 'x' in s.courses.name.foo.bar)
    @raises_exception(AttributeError, "s.courses.foo")
    def test34(self):
        select(s for s in Student if 'x' in s.courses.foo.bar)
    @raises_exception(TypeError, "Function sum() expects query or items of numeric type, got '%s' in sum(s.courses.name)" % unicode.__name__)
    def test35(self):
        select(s for s in Student if sum(s.courses.name) > 10)
    @raises_exception(TypeError, "Function sum() expects query or items of numeric type, got '%s' in sum(c.name for c in s.courses)" % unicode.__name__)
    def test36(self):
        select(s for s in Student if sum(c.name for c in s.courses) > 10)
    @raises_exception(TypeError, "Function sum() expects query or items of numeric type, got '%s' in sum(c.name for c in s.courses)" % unicode.__name__)
    def test37(self):
        select(s for s in Student if sum(c.name for c in s.courses) > 10)
    @raises_exception(TypeError, "Function avg() expects query or items of numeric type, got '%s' in avg(c.name for c in s.courses)" % unicode.__name__)
    def test38(self):
        select(s for s in Student if avg(c.name for c in s.courses) > 10 and len(s.courses) > 1)
    @raises_exception(TypeError, "strip() takes at most 1 argument (3 given)")
    def test39(self):
        select(s for s in Student if s.name.strip(1, 2, 3))
    @raises_exception(ExprEvalError,
                      "`len(1, 2) == 3` raises TypeError: len() takes exactly 1 argument (2 given)" if PYPY2 else
                      "`len(1, 2) == 3` raises TypeError: len() takes 1 positional argument but 2 were given" if PYPY else
                      "`len(1, 2) == 3` raises TypeError: len() takes exactly one argument (2 given)")
    def test40(self):
        select(s for s in Student if len(1, 2) == 3)
    @raises_exception(TypeError, "Function sum() expects query or items of numeric type, got 'Student' in sum(s for s in Student if s.group == g)")
    def test41(self):
        select(g for g in Group if sum(s for s in Student if s.group == g) > 1)
    @raises_exception(TypeError, "Function avg() expects query or items of numeric type, got 'Student' in avg(s for s in Student if s.group == g)")
    def test42(self):
        select(g for g in Group if avg(s for s in Student if s.group == g) > 1)
    @raises_exception(TypeError, "Function min() cannot be applied to type 'Student' in min(s for s in Student if s.group == g)")
    def test43(self):
        select(g for g in Group if min(s for s in Student if s.group == g) > 1)
    @raises_exception(TypeError, "Function max() cannot be applied to type 'Student' in max(s for s in Student if s.group == g)")
    def test44(self):
        select(g for g in Group if max(s for s in Student if s.group == g) > 1)
    @raises_exception(TypeError, "Attribute should be specified for 'max' aggregate function")
    def test45(self):
        max(s for s in Student)
    @raises_exception(TypeError, "Single attribute should be specified for 'max' aggregate function")
    def test46(self):
        max((s.name, s.gpa) for s in Student)
    @raises_exception(TypeError, "Attribute should be specified for 'sum' aggregate function")
    def test47(self):
        sum(s for s in Student)
    @raises_exception(TypeError, "Single attribute should be specified for 'sum' aggregate function")
    def test48(self):
        sum((s.name, s.gpa) for s in Student)
    @raises_exception(TypeError, "'sum' is valid for numeric attributes only")
    def test49(self):
        sum(s.name for s in Student)
    @raises_exception(TypeError, "Cannot compare whole JSON value, you need to select specific sub-item: s.name == {'a':'b'}")
    def test50(self):
        # cannot compare JSON value to dynamic string,
        # because a database does not provide json.dumps(s.name) functionality
        select(s for s in Student if s.name == {'a': 'b'})
    @raises_exception(IncomparableTypesError, "Incomparable types '%s' and 'int' in expression: s.name > a & 2" % unicode.__name__)
    def test51(self):
        a = 1
        select(s for s in Student if s.name > a & 2)
    @raises_exception(TypeError, "Incomparable types '%s' and 'float' in expression: s.name > 1 / a - 3" % unicode.__name__)
    def test52(self):
        a = 1
        select(s for s in Student if s.name > 1 / a - 3)
    @raises_exception(TypeError, "Incomparable types '%s' and 'int' in expression: s.name > 1 // a - 3" % unicode.__name__)
    def test53(self):
        a = 1
        select(s for s in Student if s.name > 1 // a - 3)
    @raises_exception(TypeError, "Incomparable types '%s' and 'int' in expression: s.name > -a" % unicode.__name__)
    def test54(self):
        a = 1
        select(s for s in Student if s.name > -a)
    @raises_exception(TypeError, "Incomparable types '%s' and 'list' in expression: s.name == [1, (2,)]" % unicode.__name__)
    def test55(self):
        select(s for s in Student if s.name == [1, (2,)])
    @raises_exception(TypeError, "Delete query should be applied to a single entity. Got: (s, g)")
    def test56(self):
        delete((s, g) for g in Group for s in g.students if s.gpa > 3)

if __name__ == '__main__':
    unittest.main()



settings = dict(
    provider = 'sqlite', filename = ':memory:'
    # provider='postgres', user='pony', password='pony', host='localhost', port='5432', database='pony',
    # provider='cockroach', user='root', host='localhost', port=26257, sslmode='disable', database='test'
)



from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Person(db.Entity):
    name = Required(unicode)
    passport = Optional("Passport")

class Passport(db.Entity):
    code = Required(unicode)
    person = Required("Person")

class TestOneToOne4(unittest.TestCase):
    def setUp(self):
        setup_database(db)
        with db_session:
            p1 = Person(id=1, name='John')
            Passport(id=1, code='123', person=p1)

    def tearDown(self):
        teardown_database(db)

    @raises_exception(ConstraintError, 'Cannot unlink Passport[1] from previous Person[1] object, because Passport.person attribute is required')
    @db_session
    def test1(self):
        p2 = Person(id=2, name='Mike')
        pas2 = Passport(id=2, code='456', person=p2)
        commit()
        p1 = Person.get(name='John')
        pas2.person = p1

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

from pony.orm.core import *
from pony.orm.tests import db_params

db = Database(**db_params)

class Student(db.Entity):
    _table_ = "Students"
    record = PrimaryKey(int)
    name = Required(unicode, column="fio")
    group = Required("Group")
    scholarship = Required(int, default=0)
    marks = Set("Mark")

class Group(db.Entity):
    _table_ = "Groups"
    number = PrimaryKey(str)
    department = Required(int)
    students = Set("Student")
    subjects = Set("Subject")

class Subject(db.Entity):
    _table_ = "Subjects"
    name = PrimaryKey(unicode)
    groups = Set("Group")
    marks = Set("Mark")

class Mark(db.Entity):
    _table_ = "Exams"
    student = Required(Student, column="student")
    subject = Required(Subject, column="subject")
    value = Required(int)
    PrimaryKey(student, subject)


db.generate_mapping(check_tables=False)


@db_session
def populate_db():
    Physics = Subject(name='Physics')
    Chemistry = Subject(name='Chemistry')
    Math = Subject(name='Math')

    g3132 = Group(number='3132', department=33, subjects=[ Physics, Math ])
    g4145 = Group(number='4145', department=44, subjects=[ Physics, Chemistry, Math ])
    g4146 = Group(number='4146', department=44)

    s101 = Student(record=101, name='Bob', group=g4145, scholarship=0)
    s102 = Student(record=102, name='Joe', group=g4145, scholarship=800)
    s103 = Student(record=103, name='Alex', group=g4145, scholarship=0)
    s104 = Student(record=104, name='Brad', group=g3132, scholarship=500)
    s105 = Student(record=105, name='John', group=g3132, scholarship=1000)

    Mark(student=s101, subject=Physics, value=4)
    Mark(student=s101, subject=Math, value=3)
    Mark(student=s102, subject=Chemistry, value=5)
    Mark(student=s103, subject=Physics, value=2)
    Mark(student=s103, subject=Chemistry, value=4)



from __future__ import absolute_import, print_function, division
from pony.py23compat import PY2

import unittest
from datetime import date

from pony.orm.core import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Department(db.Entity):
    number = PrimaryKey(int)
    groups = Set('Group')
    courses = Set('Course')

class Student(db.Entity):
    name = Required(unicode)
    group = Required('Group')
    scholarship = Required(int, default=0)
    picture = Optional(buffer)
    courses = Set('Course')
    grades = Set('Grade')

class Group(db.Entity):
    id = PrimaryKey(int)
    students = Set(Student)
    dept = Required(Department)
    rooms = Set('Room')

class Course(db.Entity):
    dept = Required(Department)
    name = Required(unicode)
    credits = Optional(int)
    semester = Required(int)
    PrimaryKey(name, semester)
    grades = Set('Grade')
    students = Set(Student)

class Grade(db.Entity):
    student = Required(Student)
    course = Required(Course)
    PrimaryKey(student, course)
    value = Required(str)
    date = Optional(date)
    teacher = Required('Teacher')

class Teacher(db.Entity):
    name = Required(unicode)
    grades = Set(Grade)

class Room(db.Entity):
    name = PrimaryKey(unicode)
    groups = Set(Group)


db2 = Database()

class Room2(db2.Entity):
    name = PrimaryKey(unicode)

name1 = 'S1'


class TestSQLTranslator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            d1 = Department(number=44)
            d2 = Department(number=43)
            g1 = Group(id=1, dept=d1)
            g2 = Group(id=2, dept=d2)
            s1 = Student(id=1, name='S1', group=g1, scholarship=0)
            s2 = Student(id=2, name='S2', group=g1, scholarship=100)
            s3 = Student(id=3, name='S3', group=g2, scholarship=500)
            c1 = Course(name='Math', semester=1, dept=d1)
            c2 = Course(name='Economics', semester=1, dept=d1, credits=3)
            c3 = Course(name='Physics', semester=2, dept=d2)
            t1 = Teacher(id=101, name="T1")
            t2 = Teacher(id=102, name="T2")
            Grade(student=s1, course=c1, value='C', teacher=t2, date=date(2011, 1, 1))
            Grade(student=s1, course=c3, value='A', teacher=t1, date=date(2011, 2, 1))
            Grade(student=s2, course=c2, value='B', teacher=t1)
            r1 = Room(name='Room1')
            r2 = Room(name='Room2')
            r3 = Room(name='Room3')
            g1.rooms = [r1, r2]
            g2.rooms = [r2, r3]
            c1.students.add(s1)
            c1.students.add(s2)
            c2.students.add(s2)
        setup_database(db2)
    @classmethod
    def tearDownClass(cls):
        teardown_database(db)
        teardown_database(db2)
    def setUp(self):
        rollback()
        db_session.__enter__()
    def tearDown(self):
        rollback()
        db_session.__exit__()
    def test_select1(self):
        result = set(select(s for s in Student))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})
    def test_select_param(self):
        result = select(s for s in Student if s.name == name1)[:]
        self.assertEqual(result, [Student[1]])
    def test_select_object_param(self):
        stud1 = Student[1]
        result = set(select(s for s in Student if s != stud1))
        self.assertEqual(result, {Student[2], Student[3]})
    def test_select_deref(self):
        x = 'S1'
        result = select(s for s in Student if s.name == x)[:]
        self.assertEqual(result, [Student[1]])
    def test_select_composite_key(self):
        grade1 = Grade[Student[1], Course['Physics', 2]]
        result = select(g for g in Grade if g != grade1)
        grades = [ grade.value for grade in result ]
        grades.sort()
        self.assertEqual(grades, ['B', 'C'])
    def test_function_max1(self):
        result = select(s for s in Student if max(s.grades.value) == 'C')[:]
        self.assertEqual(result, [Student[1]])
    @raises_exception(TypeError)
    def test_function_max2(self):
        grade1 = Grade[Student[1], Course['Physics', 2]]
        select(s for s in Student if max(s.grades) == grade1)
    def test_function_min(self):
        result = select(s for s in Student if min(s.grades.value) == 'B')[:]
        self.assertEqual(result, [Student[2]])
    @raises_exception(TypeError)
    def test_function_min2(self):
        grade1 = Grade[Student[1], Course['Physics', 2]]
        select(s for s in Student if min(s.grades) == grade1)
    def test_min3(self):
        d = date(2011, 1, 1)
        result = set(select(g for g in Grade if min(g.date, d) == d and g.date is not None))
        self.assertEqual(result, {Grade[Student[1], Course[u'Math', 1]], Grade[Student[1], Course[u'Physics', 2]]})
    def test_function_len1(self):
        result = select(s for s in Student if len(s.grades) == 1)[:]
        self.assertEqual(result, [Student[2]])
    def test_function_len2(self):
        result = select(s for s in Student if max(s.grades.value) == 'C')[:]
        self.assertEqual(result, [Student[1]])
    def test_function_sum1(self):
        result = select(g for g in Group if sum(g.students.scholarship) == 100)[:]
        self.assertEqual(result, [Group[1]])
    def test_function_avg1(self):
        result = select(g for g in Group if avg(g.students.scholarship) == 50)[:]
        self.assertEqual(result, [Group[1]])
    @raises_exception(TypeError)
    def test_function_sum2(self):
        select(g for g in Group if sum(g.students) == 100)
    @raises_exception(TypeError)
    def test_function_sum3(self):
        select(g for g in Group if sum(g.students.name) == 100)
    def test_function_abs(self):
        result = select(s for s in Student if abs(s.scholarship) == 100)[:]
        self.assertEqual(result, [Student[2]])
    def test_builtin_in_locals(self):
        x = max
        gen = (s.group for s in Student if x(s.grades.value) == 'C')
        result = select(gen)[:]
        self.assertEqual(result, [Group[1]])
        x = min
        result = select(gen)[:]
        self.assertEqual(result, [])
    # @raises_exception(TranslationError, "Name 'g' must be defined in query")
    # def test_name(self):
    #     select(s for s in Student for g in g.subjects)
    def test_chain1(self):
        result = set(select(g for g in Group for s in g.students if s.name.endswith('3')))
        self.assertEqual(result, {Group[2]})
    def test_chain2(self):
        result = set(select(s for g in Group if g.dept.number == 44 for s in g.students if s.name.startswith('S')))
        self.assertEqual(result, {Student[1], Student[2]})
    def test_chain_m2m(self):
        result = set(select(g for g in Group for r in g.rooms if r.name == 'Room2'))
        self.assertEqual(result, {Group[1], Group[2]})
    @raises_exception(TranslationError, 'All entities in a query must belong to the same database')
    def test_two_diagrams(self):
        select(g for g in Group for r in Room2 if r.name == 'Room2')
    def test_add_sub_mul_etc(self):
        result = select(s for s in Student if ((-s.scholarship + 200) * 10 / 5 - 100) ** 2 == 10000 or 5 == 2)[:]
        self.assertEqual(result, [Student[2]])
    def test_subscript(self):
        result = set(select(s for s in Student if s.name[1] == '2'))
        self.assertEqual(result, {Student[2]})
    def test_slice(self):
        result = set(select(s for s in Student if s.name[:1] == 'S'))
        self.assertEqual(result, {Student[3], Student[2], Student[1]})
    def test_attr_chain(self):
        s1 = Student[1]
        result = select(s for s in Student if s == s1)[:]
        self.assertEqual(result, [Student[1]])
        result = select(s for s in Student if not s == s1)[:]
        self.assertEqual(result, [Student[2], Student[3]])
        result = select(s for s in Student if s.group == s1.group)[:]
        self.assertEqual(result, [Student[1], Student[2]])
        result = select(s for s in Student if s.group.dept == s1.group.dept)[:]
        self.assertEqual(result, [Student[1], Student[2]])
    def test_list_monad1(self):
        result = select(s for s in Student if s.name in ['S1'])[:]
        self.assertEqual(result, [Student[1]])
    def test_list_monad2(self):
        result = select(s for s in Student if s.name not in ['S1', 'S2'])[:]
        self.assertEqual(result, [Student[3]])
    def test_list_monad3(self):
        grade1 = Grade[Student[1], Course['Physics', 2]]
        grade2 = Grade[Student[1], Course['Math', 1]]
        result = set(select(g for g in Grade if g in [grade1, grade2]))
        self.assertEqual(result, {grade1, grade2})
        result = set(select(g for g in Grade if g not in [grade1, grade2]))
        self.assertEqual(result, {Grade[Student[2], Course['Economics', 1]]})
    def test_tuple_monad1(self):
        n1 = 'S1'
        n2 = 'S2'
        result = select(s for s in Student if s.name in (n1, n2))[:]
        self.assertEqual(result, [Student[1], Student[2]])
    def test_None_value(self):
        result = select(s for s in Student if s.name is None)[:]
        self.assertEqual(result, [])
    def test_None_value2(self):
        result = select(s for s in Student if None == s.name)[:]
        self.assertEqual(result, [])
    def test_None_value3(self):
        n = None
        result = select(s for s in Student if s.name == n)[:]
        self.assertEqual(result, [])
    def test_None_value4(self):
        n = None
        result = select(s for s in Student if n == s.name)[:]
        self.assertEqual(result, [])
    @raises_exception(TranslationError, "External parameter 'a' cannot be used as query result")
    def test_expr1(self):
        a = 100
        result = select(a for s in Student)
    def test_expr2(self):
        result = set(select(s.group for s in Student))
        self.assertEqual(result, {Group[1], Group[2]})
    def test_numeric_binop(self):
        i = 100
        f = 2.0
        result = select(s for s in Student if s.scholarship > i + f)[:]
        self.assertEqual(result, [Student[3]])
    def test_string_const_monad(self):
        result = select(s for s in Student if len(s.name) > len('ABC'))[:]
        self.assertEqual(result, [])
    def test_numeric_to_bool1(self):
        result = set(select(s for s in Student if s.name != 'John' or s.scholarship))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})
    def test_numeric_to_bool2(self):
        result = set(select(s for s in Student if not s.scholarship))
        self.assertEqual(result, {Student[1]})
    def test_not_monad1(self):
        result = set(select(s for s in Student if not (s.scholarship > 0 and s.name != 'S1')))
        self.assertEqual(result, {Student[1]})
    def test_not_monad2(self):
        result = set(select(s for s in Student if not not (s.scholarship > 0 and s.name != 'S1')))
        self.assertEqual(result, {Student[2], Student[3]})
    def test_subquery_with_attr(self):
        result = set(select(s for s in Student if max(g.value for g in s.grades) == 'C'))
        self.assertEqual(result, {Student[1]})
    def test_query_reuse(self):
        q = select(s for s in Student if s.scholarship > 0)
        q.count()
        self.assertTrue("ORDER BY" not in db.last_sql.upper())
        objects = q[:] # should not throw exception, query can be reused
    def test_select(self):
        result = Student.select(lambda s: s.scholarship > 0)[:]
        self.assertEqual(result, [Student[2], Student[3]])
    def test_get(self):
        result = Student.get(lambda s: s.scholarship == 500)
        self.assertEqual(result, Student[3])
    def test_order_by(self):
        result = list(Student.select().order_by(Student.name))
        self.assertEqual(result, [Student[1], Student[2], Student[3]])
    def test_read_inside_query(self):
        result = set(select(s for s in Student if Group[1].dept.number == 44))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})
    def test_crud_attr_chain(self):
        result = set(select(s for s in Student if Group[1].dept.number == s.group.dept.number))
        self.assertEqual(result, {Student[1], Student[2]})
    def test_composite_key1(self):
        result = set(select(t for t in Teacher if Grade[Student[1], Course['Physics', 2]] in t.grades))
        self.assertEqual(result, {Teacher.get(name='T1')})
    def test_composite_key2(self):
        result = set(select(s for s in Student if Course['Math', 1] in s.courses))
        self.assertEqual(result, {Student[1], Student[2]})
    def test_composite_key3(self):
        result = set(select(s for s in Student if Course['Math', 1] not in s.courses))
        self.assertEqual(result, {Student[3]})
    def test_composite_key4(self):
        result = set(select(s for s in Student if len(c for c in Course if c not in s.courses) == 2))
        self.assertEqual(result, {Student[1]})
    def test_composite_key5(self):
        result = set(select(s for s in Student if not (c for c in Course if c not in s.courses)))
        self.assertEqual(result, set())
    def test_composite_key6(self):
        result = set(select(c for c in Course if c not in (c2 for s in Student for c2 in s.courses)))
        self.assertEqual(result, {Course['Physics', 2]})
    def test_composite_key7(self):
        result = set(select(c for s in Student for c in s.courses))
        self.assertEqual(result, {Course['Math', 1], Course['Economics', 1]})
    def test_contains1(self):
        s1 = Student[1]
        result = set(select(g for g in Group if s1 in g.students))
        self.assertEqual(result, {Group[1]})
    def test_contains2(self):
        s1 = Student[1]
        result = set(select(g for g in Group if s1.name in g.students.name))
        self.assertEqual(result, {Group[1]})
    def test_contains3(self):
        s1 = Student[1]
        result = set(select(g for g in Group if s1 not in g.students))
        self.assertEqual(result, {Group[2]})
    def test_contains4(self):
        s1 = Student[1]
        result = set(select(g for g in Group if s1.name not in g.students.name))
        self.assertEqual(result, {Group[2]})
    def test_buffer_monad1(self):
        try: select(s for s in Student if s.picture == buffer('abc'))
        except TypeError as e: self.assertTrue(not PY2 and str(e) == 'string argument without an encoding')
        else: self.assertTrue(PY2)
    def test_buffer_monad2(self):
        select(s for s in Student if s.picture == buffer('abc', 'ascii'))
    def test_database_monad(self):
        result = set(select(s for s in db.Student if db.Student[1] == s))
        self.assertEqual(result, {Student[1]})
    def test_duplicate_name(self):
        result = set(select(x for x in Student if x.group in (x for x in Group)))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})
    def test_hint_join1(self):
        result = set(select(s for s in Student if JOIN(max(s.courses.credits) == 3)))
        self.assertEqual(result, {Student[2]})
    def test_hint_join2(self):
        result = set(select(c for c in Course if JOIN(len(c.students) == 1)))
        self.assertEqual(result, {Course['Economics', 1]})
    def test_tuple_param(self):
        x = Student[1], Student[2]
        result = set(select(s for s in Student if s not in x))
        self.assertEqual(result, {Student[3]})
    @raises_exception(TypeError, "Expression `x` should not contain None values")
    def test_tuple_param_2(self):
        x = Student[1], None
        result = set(select(s for s in Student if s not in x))
        self.assertEqual(result, {Student[3]})
    def test_method_monad(self):
        result = set(select(s for s in Student if s not in Student.select(lambda s: s.scholarship > 0)))
        self.assertEqual(result, {Student[1]})
    def test_lambda_1(self):
        q = select(s for s in Student)
        q = q.filter(lambda s: s.name == 'S1')
        self.assertEqual(list(q), [Student[1]])
    def test_lambda_2(self):
        q = select(s for s in Student)
        q = q.filter(lambda stud: stud.name == 'S1')
        self.assertEqual(list(q), [Student[1]])
    def test_lambda_3(self):
        q = select(s for s in Student)
        q = q.filter(lambda stud: exists(x for x in Student if stud.name < x.name))
        self.assertEqual(set(q), {Student[1], Student[2]})
    def test_lambda_4(self):
        q = select(s for s in Student)
        q = q.filter(lambda stud: exists(s for s in Student if stud.name < s.name))
        self.assertEqual(set(q), {Student[1], Student[2]})
    def test_optimized_1(self):
        q = select((g, count(g.students)) for g in Group if count(g.students) > 1)
        self.assertEqual(set(q), {(Group[1], 2)})
    def test_optimized_2(self):
        q = select((s, count(s.courses)) for s in Student if count(s.courses) > 1)
        self.assertEqual(set(q), {(Student[2], 2)})
    def test_delete(self):
        q = select(g for g in Grade if g.teacher.id == 101).delete()
        q2 = select(g for g in Grade)[:]
        self.assertEqual([g.value for g in q2], ['C'])
    def test_delete_2(self):
        delete(g for g in Grade if g.teacher.id == 101)
        q2 = select(g for g in Grade)[:]
        self.assertEqual([g.value for g in q2], ['C'])
    def test_delete_3(self):
        select(g for g in Grade if g.teacher.id == 101).delete(bulk=True)
        q2 = select(g for g in Grade)[:]
        self.assertEqual([g.value for g in q2], ['C'])
    def test_delete_4(self):
        select(g for g in Grade if exists(g2 for g2 in Grade if g2.value > g.value)).delete(bulk=True)
        q2 = select(g for g in Grade)[:]
        self.assertEqual([g.value for g in q2], ['C'])
    def test_select_2(self):
        result = select(s for s in Student)[:]
        self.assertEqual(result, [Student[1], Student[2], Student[3]])
    def test_select_add(self):
        result = [None] + select(s for s in Student)[:]
        self.assertEqual(result, [None, Student[1], Student[2], Student[3]])
    def test_query_result_radd(self):
        result = select(s for s in Student)[:] + [None]
        self.assertEqual(result, [Student[1], Student[2], Student[3], None])
    def test_query_result_sort(self):
        result = select(s for s in Student)[:]
        result.sort()
        self.assertEqual(result, [Student[1], Student[2], Student[3]])
    def test_query_result_reverse(self):
        result = select(s for s in Student)[:]
        items = list(result)
        result.reverse()
        self.assertEqual(items, list(reversed(result)))
    def test_query_result_shuffle(self):
        result = select(s for s in Student)[:]
        items = set(result)
        result.shuffle()
        self.assertEqual(items, set(result))
    def test_query_result_to_list(self):
        result = select(s for s in Student)[:]
        items = result.to_list()
        self.assertTrue(type(items) is list)
    @raises_exception(TypeError, 'In order to do item assignment, cast QueryResult to list first')
    def test_query_result_setitem(self):
        result = select(s for s in Student)[:]
        result[0] = None
    @raises_exception(TypeError, 'In order to do item deletion, cast QueryResult to list first')
    def test_query_result_delitem(self):
        result = select(s for s in Student)[:]
        del result[0]
    @raises_exception(TypeError, 'In order to do +=, cast QueryResult to list first')
    def test_query_result_iadd(self):
        result = select(s for s in Student)[:]
        result += None
    @raises_exception(TypeError, 'In order to do append, cast QueryResult to list first')
    def test_query_result_append(self):
        result = select(s for s in Student)[:]
        result.append(None)
    @raises_exception(TypeError, 'In order to do clear, cast QueryResult to list first')
    def test_query_result_clear(self):
        result = select(s for s in Student)[:]
        result.clear()
    @raises_exception(TypeError, 'In order to do extend, cast QueryResult to list first')
    def test_query_result_extend(self):
        result = select(s for s in Student)[:]
        result.extend([])
    @raises_exception(TypeError, 'In order to do insert, cast QueryResult to list first')
    def test_query_result_insert(self):
        result = select(s for s in Student)[:]
        result.insert(0, None)
    @raises_exception(TypeError, 'In order to do pop, cast QueryResult to list first')
    def test_query_result_pop(self):
        result = select(s for s in Student)[:]
        result.pop()
    @raises_exception(TypeError, 'In order to do remove, cast QueryResult to list first')
    def test_query_result_remove(self):
        result = select(s for s in Student)[:]
        result.remove(None)


if __name__ == "__main__":
    unittest.main()



import unittest


from pony.orm import *
from pony.orm.tests import setup_database, teardown_database

db = Database()


class TestPost(db.Entity):
    category = Optional('TestCategory')
    name = Optional(str, default='Noname')


class TestCategory(db.Entity):
    posts = Set(TestPost)


class TransactionLockTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            cls.post = TestPost(id=1)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    __call__ = db_session(unittest.TestCase.__call__)

    def tearDown(self):
        rollback()

    def test_create(self):
        p = TestPost(id=2)
        p.flush()
        cache = db._get_cache()
        self.assertEqual(cache.immediate, True)
        self.assertEqual(cache.in_transaction, True)

    def test_update(self):
        p = TestPost[self.post.id]
        p.name = 'Trash'
        p.flush()
        cache = db._get_cache()
        self.assertEqual(cache.immediate, True)
        self.assertEqual(cache.in_transaction, True)

    def test_delete(self):
        p = TestPost[self.post.id]
        p.delete()
        flush()
        cache = db._get_cache()
        self.assertEqual(cache.immediate, True)
        self.assertEqual(cache.in_transaction, True)



import unittest

from pony.orm.core import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Group(db.Entity):
    students = Set('Student')

class Student(db.Entity):
    first_name = Required(str)
    last_name = Required(str)
    login = Optional(str, nullable=True)
    graduated = Optional(bool, default=False)
    group = Required(Group)
    passport = Optional('Passport', column='passport')

class Passport(db.Entity):
    student = Optional(Student)


class TestExists(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            g1 = Group(id=1)
            g2 = Group(id=2)

            p = Passport(id=1)

            Student(id=1, first_name='Mashu', last_name='Kyrielight', login='Shielder', group=g1)
            Student(id=2, first_name='Okita', last_name='Souji', login='Sakura', group=g1)
            Student(id=3, first_name='Francis', last_name='Drake', group=g2, graduated=True)
            Student(id=4, first_name='Oda', last_name='Nobunaga', group=g2, graduated=True)
            Student(id=5, first_name='William', last_name='Shakespeare', group=g2, graduated=True, passport=p)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_1(self):
        q = select(g for g in Group if exists(s.login for s in g.students))[:]
        self.assertEqual(q[0], Group[1])

    def test_2(self):
        q = select(g for g in Group if exists(s.graduated for s in g.students))[:]
        self.assertEqual(q[0], Group[2])

    def test_3(self):
        q = select(s for s in Student if
                   exists(len(s2.first_name) == len(s.first_name) and s != s2 for s2 in Student))[:]
        self.assertEqual(set(q), {Student[1], Student[2], Student[3], Student[5]})

    def test_4(self):
        q = select(g for g in Group if not exists(not s.graduated for s in g.students))[:]
        self.assertEqual(q[0], Group[2])

    def test_5(self):
        q = select(g for g in Group if exists(s for s in g.students))[:]
        self.assertEqual(set(q), {Group[1], Group[2]})

    def test_6(self):
        q = select(g for g in Group if exists(s.login for s in g.students if s.first_name != 'Okita') and g.id != 10)[:]
        self.assertEqual(q[0], Group[1])

    def test_7(self):
        q = select(g for g in Group if exists(s.passport for s in g.students))[:]
        self.assertEqual(q[0], Group[2])


from __future__ import absolute_import, print_function, division

import unittest
from datetime import date

from pony.orm.core import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Department(db.Entity):
    number = PrimaryKey(int, auto=True)
    name = Required(unicode, unique=True)
    groups = Set("Group")
    courses = Set("Course")

class Group(db.Entity):
    number = PrimaryKey(int)
    major = Required(unicode)
    dept = Required("Department")
    students = Set("Student")

class Course(db.Entity):
    name = Required(unicode)
    semester = Required(int)
    lect_hours = Required(int)
    lab_hours = Required(int)
    credits = Required(int)
    dept = Required(Department)
    students = Set("Student")
    PrimaryKey(name, semester)

class Student(db.Entity):
    id = PrimaryKey(int, auto=True)
    name = Required(unicode)
    dob = Required(date)
    tel = Optional(str)
    picture = Optional(buffer, lazy=True)
    gpa = Required(float, default=0)
    phd = Optional(bool)
    group = Required(Group)
    courses = Set(Course)

class TestListMonad(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            d1 = Department(number=1, name="Department of Computer Science")
            d2 = Department(number=2, name="Department of Mathematical Sciences")
            d3 = Department(number=3, name="Department of Applied Physics")

            c1 = Course(name="Web Design", semester=1, dept=d1,
                        lect_hours=30, lab_hours=30, credits=3)
            c2 = Course(name="Data Structures and Algorithms", semester=3, dept=d1,
                        lect_hours=40, lab_hours=20, credits=4)

            c3 = Course(name="Linear Algebra", semester=1, dept=d2,
                        lect_hours=30, lab_hours=30, credits=4)
            c4 = Course(name="Statistical Methods", semester=2, dept=d2,
                        lect_hours=50, lab_hours=25, credits=5)

            c5 = Course(name="Thermodynamics", semester=2, dept=d3,
                        lect_hours=25, lab_hours=40, credits=4)
            c6 = Course(name="Quantum Mechanics", semester=3, dept=d3,
                        lect_hours=40, lab_hours=30, credits=5)

            g101 = Group(number=101, major='B.E. in Computer Engineering', dept=d1)
            g102 = Group(number=102, major='B.S./M.S. in Computer Science', dept=d2)
            g103 = Group(number=103, major='B.S. in Applied Mathematics and Statistics', dept=d2)
            g104 = Group(number=104, major='B.S./M.S. in Pure Mathematics', dept=d2)
            g105 = Group(number=105, major='B.E in Electronics', dept=d3)
            g106 = Group(number=106, major='B.S./M.S. in Nuclear Engineering', dept=d3)

            Student(id=1, name='John Smith', dob=date(1991, 3, 20), tel='123-456', gpa=3, group=g101, phd=True,
                    courses=[c1, c2, c4, c6])
            Student(id=2, name='Matthew Reed', dob=date(1990, 11, 26), gpa=3.5, group=g101, phd=True,
                    courses=[c1, c3, c4, c5])
            Student(id=3, name='Chuan Qin', dob=date(1989, 2, 5), gpa=4, group=g101,
                    courses=[c3, c5, c6])
            Student(id=4, name='Rebecca Lawson', dob=date(1990, 4, 18), tel='234-567', gpa=3.3, group=g102,
                    courses=[c1, c4, c5, c6])
            Student(id=5, name='Maria Ionescu', dob=date(1991, 4, 23), gpa=3.9, group=g102,
                    courses=[c1, c2, c4, c6])
            Student(id=6, name='Oliver Blakey', dob=date(1990, 9, 8), gpa=3.1, group=g102,
                    courses=[c1, c2, c5])
            Student(id=7, name='Jing Xia', dob=date(1988, 12, 30), gpa=3.2, group=g102,
                    courses=[c1, c3, c5, c6])

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_in_simple(self):
        q = select(s.id for s in Student if s.name in ('John Smith', 'Matthew Reed'))
        self.assertEqual(set(q), {1, 2})

    def test_not_in_simple(self):
        q = select(s.id for s in Student if s.name not in ('John Smith', 'Matthew Reed'))
        self.assertEqual(set(q), {3, 4, 5, 6, 7})

    def test_in_composite(self):
        q = select(c.name for c in Course if (c.name, c.semester) in [
            ("Web Design", 1), ("Thermodynamics", 2), ("Theology", 3)
        ])
        self.assertEqual(set(q), {"Web Design", "Thermodynamics"})

    def test_not_in_composite(self):
        q = select(c.name for c in Course if (c.name, c.semester) not in [
            ("Web Design", 1), ("Thermodynamics", 2), ("Theology", 3)
        ])
        self.assertEqual(set(q), {
            "Data Structures and Algorithms", "Linear Algebra",
            "Statistical Methods", "Quantum Mechanics"
        })

    def test_in_simple_object(self):
        s1, s2 = Student[1], Student[2]
        q = select(s.id for s in Student if s in (s1, s2))
        self.assertEqual(set(q), {1, 2})

    def test_not_in_simple_object(self):
        s1, s2 = Student[1], Student[2]
        q = select(s.id for s in Student if s not in (s1, s2))
        self.assertEqual(set(q), {3, 4, 5, 6, 7})

    def test_in_composite_object(self):
        c1, c2 = Course["Web Design", 1], Course["Thermodynamics", 2]
        q = select(c.name for c in Course if c in (c1, c2))
        self.assertEqual(set(q), {"Web Design", "Thermodynamics"})

    def test_not_in_composite_object(self):
        c1, c2 = Course["Web Design", 1], Course["Thermodynamics", 2]
        q = select(c.name for c in Course if c not in (c1, c2))
        self.assertEqual(set(q), {
            "Data Structures and Algorithms", "Linear Algebra",
            "Statistical Methods", "Quantum Mechanics"
        })


if __name__ == "__main__":
    unittest.main()



import unittest

from pony.orm import *
from pony.orm.tests.testutils import raises_exception
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Person(db.Entity):
    name = Required(str)
    tel = Optional(str)


class TestAutostrip(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    @db_session
    def test_1(self):
        p = Person(name=' John  ', tel=' ')
        p.flush()
        self.assertEqual(p.name, 'John')
        self.assertEqual(p.tel, '')

    @raises_exception(ValueError, 'Attribute Person.name is required')
    @db_session
    def test_2(self):
        p = Person(name=' ')

if __name__ == '__main__':
    unittest.main()



from datetime import date
from decimal import Decimal

import unittest

from pony.orm import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database, only_for

db = Database()


class Person(db.Entity):
    id = PrimaryKey(int, auto=True)
    name = Required(str)
    dob = Optional(date)
    ssn = Required(str, unique=True)


class Student(Person):
    group = Required("Group")
    mentor = Optional("Teacher")
    attend_courses = Set("Course")


class Teacher(Person):
    teach_courses = Set("Course")
    apprentices = Set("Student")
    salary = Required(Decimal)


class Assistant(Student, Teacher):
    pass


class Professor(Teacher):
    position = Required(str)


class Group(db.Entity):
    number = PrimaryKey(int)
    students = Set("Student")


class Course(db.Entity):
    name = Required(str)
    semester = Required(int)
    students = Set(Student)
    teachers = Set(Teacher)
    PrimaryKey(name, semester)


class TestVolatile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            p = Person(name='Person1', ssn='SSN1')
            g = Group(number=123)
            prof = Professor(name='Professor1', salary=1000, position='position1', ssn='SSN5')
            a1 = Assistant(name='Assistant1', group=g, salary=100, ssn='SSN4', mentor=prof)
            a2 = Assistant(name='Assistant2', group=g, salary=200, ssn='SSN6', mentor=prof)
            s1 = Student(name='Student1', group=g, ssn='SSN2', mentor=a1)
            s2 = Student(name='Student2', group=g, ssn='SSN3')

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    @db_session
    def test_1(self):
        q = select(p.name for p in Person if isinstance(p, Student))
        self.assertEqual(set(q), {'Student1', 'Student2', 'Assistant1', 'Assistant2'})

    @db_session
    def test_2(self):
        q = select(p.name for p in Person if not isinstance(p, Student))
        self.assertEqual(set(q), {'Person1', 'Professor1'})

    @db_session
    def test_3(self):
        q = select(p.name for p in Student if isinstance(p, Professor))
        self.assertEqual(set(q), set())

    @db_session
    def test_4(self):
        q = select(p.name for p in Person if not isinstance(p, Person))
        self.assertEqual(set(q), set())

    @db_session
    def test_5(self):
        q = select(p.name for p in Person if isinstance(p, (Student, Teacher)))
        self.assertEqual(set(q), {'Student1', 'Student2', 'Assistant1', 'Assistant2', 'Professor1'})

    @db_session
    def test_6(self):
        q = select(p.name for p in Person if isinstance(p, Student) and isinstance(p, Teacher))
        self.assertEqual(set(q), {'Assistant1', 'Assistant2'})

    @db_session
    def test_7(self):
        q = select(p.name for p in Person
                   if (isinstance(p, Student) and p.ssn == 'SSN2')
                   or (isinstance(p, Professor) and p.salary > 500))
        self.assertEqual(set(q), {'Student1', 'Professor1'})

    @db_session
    def test_8(self):
        q = select(p.name for p in Person if isinstance(p, Person))
        self.assertEqual(set(q), {'Person1', 'Student1', 'Student2', 'Assistant1', 'Assistant2', 'Professor1'})

    @db_session
    def test_9(self):
        q = select(g.number for g in Group if isinstance(g, Group))
        self.assertEqual(set(q), {123})


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function
from pony.py23compat import PY2, imap, basestring, unicode, pickle, iteritems

import io, re, os.path, sys, inspect, types, warnings

from datetime import datetime
from itertools import count as _count
from inspect import isfunction
from time import strptime
from collections import defaultdict
from functools import update_wrapper, wraps
from xml.etree import cElementTree
from copy import deepcopy

import pony
from pony import options

from pony.thirdparty.compiler import ast
from pony.thirdparty.decorator import decorator as _decorator

if pony.MODE.startswith('GAE-'): localbase = object
else: from threading import local as localbase


class PonyDeprecationWarning(DeprecationWarning):
    pass

def deprecated(stacklevel, message):
    warnings.warn(message, PonyDeprecationWarning, stacklevel)

warnings.simplefilter('once', PonyDeprecationWarning)

def _improved_decorator(caller, func):
    if isfunction(func):
        return _decorator(caller, func)
    def pony_wrapper(*args, **kwargs):
        return caller(func, *args, **kwargs)
    return pony_wrapper

def decorator(caller, func=None):
    if func is not None:
        return _improved_decorator(caller, func)
    def new_decorator(func):
        return _improved_decorator(caller, func)
    if isfunction(caller):
        update_wrapper(new_decorator, caller)
    return new_decorator

def decorator_with_params(dec):
    def parameterized_decorator(*args, **kwargs):
        if len(args) == 1 and isfunction(args[0]) and not kwargs:
            return decorator(dec(), args[0])
        return decorator(dec(*args, **kwargs))
    return parameterized_decorator

@decorator
def cut_traceback(func, *args, **kwargs):
    if not options.CUT_TRACEBACK:
        return func(*args, **kwargs)

    try: return func(*args, **kwargs)
    except AssertionError: raise
    except Exception:
        exc_type, exc, tb = sys.exc_info()
        full_tb = tb
        last_pony_tb = None
        try:
            while tb.tb_next:
                module_name = tb.tb_frame.f_globals['__name__']
                if module_name == 'pony' or (module_name is not None  # may be None during import
                                             and module_name.startswith('pony.')):
                    last_pony_tb = tb
                tb = tb.tb_next
            if last_pony_tb is None: raise
            module_name = tb.tb_frame.f_globals.get('__name__') or ''
            if module_name.startswith('pony.utils') and tb.tb_frame.f_code.co_name == 'throw':
                reraise(exc_type, exc, last_pony_tb)
            reraise(exc_type, exc, full_tb)
        finally:
            del exc, full_tb, tb, last_pony_tb

cut_traceback_depth = 2

if pony.MODE != 'INTERACTIVE':
    cut_traceback_depth = 0
    def cut_traceback(func):
        return func

if PY2:
    exec('''def reraise(exc_type, exc, tb):
    try: raise exc_type, exc, tb
    finally: del tb''')
else:
    def reraise(exc_type, exc, tb):
        try: raise exc.with_traceback(tb)
        finally: del exc, tb

def throw(exc_type, *args, **kwargs):
    if isinstance(exc_type, Exception):
        assert not args and not kwargs
        exc = exc_type
    else: exc = exc_type(*args, **kwargs)
    exc.__cause__ = None
    try:
        if not (pony.MODE == 'INTERACTIVE' and options.CUT_TRACEBACK):
            raise exc
        else:
            raise exc  # Set "pony.options.CUT_TRACEBACK = False" to see full traceback
    finally: del exc

def truncate_repr(s, max_len=100):
    s = repr(s)
    return s if len(s) <= max_len else s[:max_len-3] + '...'

codeobjects = {}

def get_codeobject_id(codeobject):
    codeobject_id = id(codeobject)
    if codeobject_id not in codeobjects:
        codeobjects[codeobject_id] = codeobject
    return codeobject_id

lambda_args_cache = {}

def get_lambda_args(func):
    if type(func) is types.FunctionType:
        codeobject = func.func_code if PY2 else func.__code__
        cache_key = get_codeobject_id(codeobject)
    elif isinstance(func, ast.Lambda):
        cache_key = func
    else: assert False  # pragma: no cover

    names = lambda_args_cache.get(cache_key)
    if names is not None: return names

    if type(func) is types.FunctionType:
        if hasattr(inspect, 'signature'):
            names, argsname, kwname, defaults = [], None, None, []
            for p in inspect.signature(func).parameters.values():
                if p.default is not p.empty:
                    defaults.append(p.default)

                if p.kind == p.POSITIONAL_OR_KEYWORD:
                    names.append(p.name)
                elif p.kind == p.VAR_POSITIONAL:
                    argsname = p.name
                elif p.kind == p.VAR_KEYWORD:
                    kwname = p.name
                elif p.kind == p.POSITIONAL_ONLY:
                    throw(TypeError, 'Positional-only arguments like %s are not supported' % p.name)
                elif p.kind == p.KEYWORD_ONLY:
                    throw(TypeError, 'Keyword-only arguments like %s are not supported' % p.name)
                else: assert False
        else:
            names, argsname, kwname, defaults = inspect.getargspec(func)
    elif isinstance(func, ast.Lambda):
        names = func.argnames
        if func.kwargs: names, kwname = names[:-1], names[-1]
        else: kwname = None
        if func.varargs: names, argsname = names[:-1], names[-1]
        else: argsname = None
        defaults = func.defaults
    else: assert False  # pragma: no cover
    if argsname: throw(TypeError, '*%s is not supported' % argsname)
    if kwname: throw(TypeError, '**%s is not supported' % kwname)
    if defaults: throw(TypeError, 'Defaults are not supported')

    lambda_args_cache[cache_key] = names
    return names

def error_method(*args, **kwargs):
    raise TypeError()

_ident_re = re.compile(r'^[A-Za-z_]\w*\Z')

# is_ident = ident_re.match
def is_ident(string):
    'is_ident(string) -> bool'
    return bool(_ident_re.match(string))

_name_parts_re = re.compile(r'''
            [A-Z][A-Z0-9]+(?![a-z]) # ACRONYM
        |   [A-Z][a-z]*             # Capitalized or single capital
        |   [a-z]+                  # all-lowercase
        |   [0-9]+                  # numbers
        |   _+                      # underscores
        ''', re.VERBOSE)

def split_name(name):
    "split_name('Some_FUNNYName') -> ['Some', 'FUNNY', 'Name']"
    if not _ident_re.match(name):
        raise ValueError('Name is not correct Python identifier')
    list = _name_parts_re.findall(name)
    if not (list[0].strip('_') and list[-1].strip('_')):
        raise ValueError('Name must not starting or ending with underscores')
    return [ s for s in list if s.strip('_') ]

def uppercase_name(name):
    "uppercase_name('Some_FUNNYName') -> 'SOME_FUNNY_NAME'"
    return '_'.join(s.upper() for s in split_name(name))

def lowercase_name(name):
    "uppercase_name('Some_FUNNYName') -> 'some_funny_name'"
    return '_'.join(s.lower() for s in split_name(name))

def camelcase_name(name):
    "uppercase_name('Some_FUNNYName') -> 'SomeFunnyName'"
    return ''.join(s.capitalize() for s in split_name(name))

def mixedcase_name(name):
    "mixedcase_name('Some_FUNNYName') -> 'someFunnyName'"
    list = split_name(name)
    return list[0].lower() + ''.join(s.capitalize() for s in list[1:])

def import_module(name):
    "import_module('a.b.c') -> <module a.b.c>"
    mod = sys.modules.get(name)
    if mod is not None: return mod
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]: mod = getattr(mod, comp)
    return mod

if sys.platform == 'win32':
      _absolute_re = re.compile(r'^(?:[A-Za-z]:)?[\\/]')
else: _absolute_re = re.compile(r'^/')

def is_absolute_path(filename):
    return bool(_absolute_re.match(filename))

def absolutize_path(filename, frame_depth):
    if is_absolute_path(filename): return filename
    code_filename = sys._getframe(frame_depth+1).f_code.co_filename
    if not is_absolute_path(code_filename):
        if code_filename.startswith('<') and code_filename.endswith('>'):
            if pony.MODE == 'INTERACTIVE': raise ValueError(
                'When in interactive mode, please provide absolute file path. Got: %r' % filename)
            raise EnvironmentError('Unexpected module filename, which is not absolute file path: %r' % code_filename)
    code_path = os.path.dirname(code_filename)
    return os.path.join(code_path, filename)

def current_timestamp():
    return datetime2timestamp(datetime.now())

def datetime2timestamp(d):
    result = d.isoformat(' ')
    if len(result) == 19: return result + '.000000'
    return result

def timestamp2datetime(t):
    time_tuple = strptime(t[:19], '%Y-%m-%d %H:%M:%S')
    microseconds = int((t[20:26] + '000000')[:6])
    return datetime(*(time_tuple[:6] + (microseconds,)))

expr1_re = re.compile(r'''
        ([A-Za-z_]\w*)  # identifier (group 1)
    |   ([(])           # open parenthesis (group 2)
    ''', re.VERBOSE)

expr2_re = re.compile(r'''
     \s*(?:
            (;)                 # semicolon (group 1)
        |   (\.\s*[A-Za-z_]\w*) # dot + identifier (group 2)
        |   ([([])              # open parenthesis or braces (group 3)
        )
    ''', re.VERBOSE)

expr3_re = re.compile(r"""
        [()[\]]                   # parenthesis or braces (group 1)
    |   '''(?:[^\\]|\\.)*?'''     # '''triple-quoted string'''
    |   \"""(?:[^\\]|\\.)*?\"""   # \"""triple-quoted string\"""
    |   '(?:[^'\\]|\\.)*?'        # 'string'
    |   "(?:[^"\\]|\\.)*?"        # "string"
    """, re.VERBOSE)

def parse_expr(s, pos=0):
    z = 0
    match = expr1_re.match(s, pos)
    if match is None: raise ValueError()
    start = pos
    i = match.lastindex
    if i == 1: pos = match.end()  # identifier
    elif i == 2: z = 2  # "("
    else: assert False  # pragma: no cover
    while True:
        match = expr2_re.match(s, pos)
        if match is None: return s[start:pos], z==1
        pos = match.end()
        i = match.lastindex
        if i == 1: return s[start:pos], False  # ";" - explicit end of expression
        elif i == 2: z = 2  # .identifier
        elif i == 3:  # "(" or "["
            pos = match.end()
            counter = 1
            open = match.group(i)
            if open == '(': close = ')'
            elif open == '[': close = ']'; z = 2
            else: assert False  # pragma: no cover
            while True:
                match = expr3_re.search(s, pos)
                if match is None: raise ValueError()
                pos = match.end()
                x = match.group()
                if x == open: counter += 1
                elif x == close:
                    counter -= 1
                    if not counter: z += 1; break
        else: assert False  # pragma: no cover

def tostring(x):
    if isinstance(x, basestring): return x
    if hasattr(x, '__unicode__'):
        try: return unicode(x)
        except: pass
    if hasattr(x, 'makeelement'): return cElementTree.tostring(x)
    try: return str(x)
    except: pass
    try: return repr(x)
    except: pass
    if type(x) == types.InstanceType: return '<%s instance at 0x%X>' % (x.__class__.__name__)
    return '<%s object at 0x%X>' % (x.__class__.__name__)

def strjoin(sep, strings, source_encoding='ascii', dest_encoding=None):
    "Can join mix of unicode and byte strings in different encodings"
    strings = list(strings)
    try: return sep.join(strings)
    except UnicodeDecodeError: pass
    for i, s in enumerate(strings):
        if isinstance(s, str):
            strings[i] = s.decode(source_encoding, 'replace').replace(u'\ufffd', '?')
    result = sep.join(strings)
    if dest_encoding is None: return result
    return result.encode(dest_encoding, 'replace')

def count(*args, **kwargs):
    if kwargs: return _count(*args, **kwargs)
    if len(args) != 1: return _count(*args)
    arg = args[0]
    if hasattr(arg, 'count'): return arg.count()
    try: it = iter(arg)
    except TypeError: return _count(arg)
    return len(set(it))

def avg(iter):
    count = 0
    sum = 0.0
    for elem in iter:
        if elem is None: continue
        sum += elem
        count += 1
    if not count: return None
    return sum / count

def group_concat(items, sep=','):
    if items is None:
        return None
    return str(sep).join(str(item) for item in items)

def coalesce(*args):
    for arg in args:
        if arg is not None:
            return arg
    return None

def distinct(iter):
    d = defaultdict(int)
    for item in iter:
        d[item] = d[item] + 1
    return d

def concat(*args):
    return ''.join(tostring(arg) for arg in args)

def between(x, a, b):
    return a <= x <= b

def is_utf8(encoding):
    return encoding.upper().replace('_', '').replace('-', '') in ('UTF8', 'UTF', 'U8')

def _persistent_id(obj):
    if obj is Ellipsis:
        return "Ellipsis"

def _persistent_load(persid):
    if persid == "Ellipsis":
        return Ellipsis
    raise pickle.UnpicklingError("unsupported persistent object")

def pickle_ast(val):
    pickled = io.BytesIO()
    pickler = pickle.Pickler(pickled)
    pickler.persistent_id = _persistent_id
    pickler.dump(val)
    return pickled

def unpickle_ast(pickled):
    pickled.seek(0)
    unpickler = pickle.Unpickler(pickled)
    unpickler.persistent_load = _persistent_load
    return unpickler.load()

def copy_ast(tree):
    return unpickle_ast(pickle_ast(tree))

def _hashable_wrap(func):
    @wraps(func, assigned=('__name__', '__doc__'))
    def new_func(self, *args, **kwargs):
        if getattr(self, '_hash', None) is not None:
            assert False, 'Cannot mutate HashableDict instance after the hash value is calculated'
        return func(self, *args, **kwargs)
    return new_func

class HashableDict(dict):
    def __hash__(self):
        result = getattr(self, '_hash', None)
        if result is None:
            result = 0
            for key, value in self.items():
                result ^= hash(key)
                result ^= hash(value)
            self._hash = result
        return result
    def __deepcopy__(self, memo):
        if getattr(self, '_hash', None) is not None:
            return self
        return HashableDict({deepcopy(key, memo): deepcopy(value, memo)
                            for key, value in iteritems(self)})
    __setitem__ = _hashable_wrap(dict.__setitem__)
    __delitem__ = _hashable_wrap(dict.__delitem__)
    clear = _hashable_wrap(dict.clear)
    pop = _hashable_wrap(dict.pop)
    popitem = _hashable_wrap(dict.popitem)
    setdefault = _hashable_wrap(dict.setdefault)
    update = _hashable_wrap(dict.update)

def deref_proxy(value):
    t = type(value)
    if t.__name__ == 'LocalProxy' and '_get_current_object' in t.__dict__:
        # Flask local proxy
        value = value._get_current_object()
    elif t.__name__ == 'EntityProxy':
        # Pony proxy
        value = value._get_object()

    return value

def deduplicate(value, deduplication_cache):
    t = type(value)
    try:
        return deduplication_cache[t].setdefault(value, value)
    except:
        return value



from __future__ import absolute_import, print_function, division
from pony.py23compat import PY2

import unittest

from pony.orm.tests.testutils import raises_exception
from pony.orm.tests.model1 import *
from pony.orm.tests import setup_database, teardown_database


class TestCollections(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        populate_db()

    @classmethod
    def tearDownClass(cls):
        db.drop_all_tables(with_all_data=True)

    @db_session
    def test_setwrapper_len(self):
        g = Group.get(number='4145')
        self.assertTrue(len(g.students) == 3)

    @db_session
    def test_setwrapper_nonzero(self):
        g = Group.get(number='4145')
        self.assertTrue(bool(g.students) == True)
        self.assertTrue(len(g.students) == 3)

    @db_session
    @raises_exception(TypeError, 'Collection attribute Group.students cannot be specified as search criteria')
    def test_get_by_collection_error(self):
        Group.get(students=[])

    @db_session
    def test_collection_create_one2many_1(self):
        g = Group['3132']
        g.students.create(record=106, name='Mike', scholarship=200)
        flush()
        self.assertEqual(len(g.students), 3)
        rollback()

    @raises_exception(TypeError, "When using Group.students.create(), "
                                 "'group' attribute should not be passed explicitly")
    @db_session
    def test_collection_create_one2many_2(self):
        g = Group['3132']
        g.students.create(record=106, name='Mike', scholarship=200, group=g)

    @raises_exception(TransactionIntegrityError, "Object Student[105] cannot be stored in the database...")
    @db_session
    def test_collection_create_one2many_3(self):
        g = Group['3132']
        g.students.create(record=105, name='Mike', scholarship=200)

    @db_session
    def test_collection_create_many2many_1(self):
        g = Group['3132']
        g.subjects.create(name='Biology')
        flush()
        self.assertEqual(len(g.subjects), 3)
        rollback()

    @raises_exception(TypeError, "When using Group.subjects.create(), "
                                 "'groups' attribute should not be passed explicitly")
    @db_session
    def test_collection_create_many2many_2(self):
        g = Group['3132']
        g.subjects.create(name='Biology', groups=[g])

    @raises_exception(TransactionIntegrityError,
                      "Object Subject[u'Math'] cannot be stored in the database..." if PY2 else
                      "Object Subject['Math'] cannot be stored in the database...")
    @db_session
    def test_collection_create_many2many_3(self):
        g = Group['3132']
        g.subjects.create(name='Math')

# replace collection items when the old ones are not fully loaded
##>>> from pony.examples.orm.students01.model import *
##>>> s1 = Student[101]
##>>> g = s1.group
##>>> g.__dict__[Group.students].is_fully_loaded
##False
##>>> s2 = Student[104]
##>>> g.students = [s2]
##>>>

# replace collection items when the old ones are not loaded
##>>> from pony.examples.orm.students01.model import *
##>>> g = Group[4145]
##>>> Group.students not in g.__dict__
##True
##>>> s2 = Student[104]
##>>> g.students = [s2]


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import types
import pony.orm.core, pony.options

pony.options.CUT_TRACEBACK = False
pony.orm.core.sql_debug(False)


def _load_env():
    settings_filename = os.environ.get('pony_test_db')
    if settings_filename is None:
        print('use default sqlite provider')
        return dict(provider='sqlite', filename=':memory:')
    with open(settings_filename, 'r') as f:
        content = f.read()

    config = {}
    exec(content, config)
    settings = config.get('settings')
    if settings is None or not isinstance(settings, dict):
        raise ValueError('Incorrect settings pony test db file contents')
    provider = settings.get('provider')
    if provider is None:
        raise ValueError('Incorrect settings pony test db file contents: provider was not specified')
    print('use provider %s' % provider)
    return settings


db_params = _load_env()


def setup_database(db):
    if db.provider is None:
        db.bind(**db_params)
    if db.schema is None:
        db.generate_mapping(check_tables=False)
    db.drop_all_tables(with_all_data=True)
    db.create_tables()


def teardown_database(db):
    if db.schema:
        db.drop_all_tables(with_all_data=True)
    db.disconnect()


def only_for(providers):
    if not isinstance(providers, (list, tuple)):
        providers = [providers]
    def decorator(x):
        if isinstance(x, type) and issubclass(x, unittest.TestCase):
            @classmethod
            def setUpClass(cls):
                raise unittest.SkipTest('%s tests implemented only for %s provider%s' % (
                    cls.__name__, ', '.join(providers), '' if len(providers) < 2 else 's'
                ))
            if db_params['provider'] not in providers:
                x.setUpClass = setUpClass
            result = x
        elif isinstance(x, types.FunctionType):
            def new_test_func(self):
                if db_params['provider'] not in providers:
                    raise unittest.SkipTest('%s test implemented only for %s provider%s' % (
                        x.__name__, ', '.join(providers), '' if len(providers) < 2 else 's'
                    ))
                return x(self)
            result = new_test_func
        else:
            raise TypeError
        return result
    return decorator



from __future__ import absolute_import, print_function, division
from pony.py23compat import PYPY2

import unittest
from datetime import date

from pony.orm import *
from pony.orm.tests.testutils import raises_exception
from pony.orm.tests import setup_database, teardown_database, only_for

db = Database()


class Person(db.Entity):
    id = PrimaryKey(int)
    name = Required(str)
    age = Required(int)
    dob = Required(date)


@only_for('sqlite')
class TestRawSQL(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            Person(id=1, name='John', age=30, dob=date(1985, 1, 1))
            Person(id=2, name='Mike', age=32, dob=date(1983, 5, 20))
            Person(id=3, name='Mary', age=20, dob=date(1995, 2, 15))

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    @db_session
    def test_1(self):
        # raw_sql result can be treated as a logical expression
        persons = select(p for p in Person if raw_sql('abs("p"."age") > 25'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_2(self):
        # raw_sql result can be used for comparison
        persons = select(p for p in Person if raw_sql('abs("p"."age")') > 25)[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_3(self):
        # raw_sql can accept $parameters
        x = 25
        persons = select(p for p in Person if raw_sql('abs("p"."age") > $x'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_4(self):
        # dynamic raw_sql content (1)
        x = 1
        s = 'p.id > $x'
        persons = select(p for p in Person if raw_sql(s))[:]
        self.assertEqual(set(persons), {Person[2], Person[3]})

    @db_session
    def test_5(self):
        # dynamic raw_sql content (2)
        x = 1
        cond = raw_sql('p.id > $x')
        persons = select(p for p in Person if cond)[:]
        self.assertEqual(set(persons), {Person[2], Person[3]})

    @db_session
    def test_6(self):
        # correct converter should be applied to raw_sql parameter type
        x = date(1990, 1, 1)
        persons = select(p for p in Person if raw_sql('p.dob < $x'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_7(self):
        # raw_sql argument may be complex expression (1)
        x = 10
        y = 15
        persons = select(p for p in Person if raw_sql('p.age > $(x + y)'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_8(self):
        # raw_sql argument may be complex expression (2)
        persons = select(p for p in Person if raw_sql('p.dob < $date.today()'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2], Person[3]})

    @db_session
    def test_9(self):
        # using raw_sql in the expression part of the generator
        names = select(raw_sql('UPPER(p.name)') for p in Person)[:]
        self.assertEqual(set(names), {'JOHN', 'MIKE', 'MARY'})

    @db_session
    def test_10(self):
        # raw_sql does not know result type and cannot appy correct type converter automatically
        dates = select(raw_sql('(p.dob)') for p in Person).order_by(lambda: p.id)[:]
        self.assertEqual(dates, ['1985-01-01', '1983-05-20', '1995-02-15'])

    @db_session
    def test_11(self):
        # it is possible to specify raw_sql type manually
        dates = select(raw_sql('(p.dob)', result_type=date) for p in Person).order_by(lambda: p.id)[:]
        self.assertEqual(dates, [date(1985, 1, 1), date(1983, 5, 20), date(1995, 2, 15)])

    @db_session
    def test_12(self):
        # raw_sql can be used in lambdas
        x = 25
        persons = Person.select(lambda p: p.age > raw_sql('$x'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_13(self):
        # raw_sql in filter()
        x = 25
        persons = select(p for p in Person).filter(lambda p: p.age > raw_sql('$x'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_14(self):
        # raw_sql in filter() without using lambda
        x = 25
        persons = Person.select().filter(raw_sql('p.age > $x'))[:]
        self.assertEqual(set(persons), {Person[1], Person[2]})

    @db_session
    def test_15(self):
        # several raw_sql expressions in a single query
        x = '123'
        y = 'John'
        persons = Person.select(lambda p: raw_sql("UPPER(p.name) || $x") == raw_sql("UPPER($y || '123')"))[:]
        self.assertEqual(set(persons), {Person[1]})

    @db_session
    def test_16(self):
        # the same param name can be used several times with different types & values
        x = 10
        y = 31
        q = select(p for p in Person if p.age > x and p.age < raw_sql('$y'))
        x = date(1980, 1, 1)
        y = 'j'
        q = q.filter(lambda p: p.dob > x and p.name.startswith(raw_sql('UPPER($y)')))
        persons = q[:]
        self.assertEqual(set(persons), {Person[1]})

    @db_session
    def test_17(self):
        # raw_sql in order_by() section
        x = 9
        persons = Person.select().order_by(lambda p: raw_sql('SUBSTR(p.dob, $x)'))[:]
        self.assertEqual(persons, [Person[1], Person[3], Person[2]])

    @db_session
    def test_18(self):
        # raw_sql in order_by() section without using lambda
        x = 9
        persons = Person.select().order_by(raw_sql('SUBSTR(p.dob, $x)'))[:]
        self.assertEqual(persons, [Person[1], Person[3], Person[2]])

    @db_session
    @raises_exception(TranslationError, "Expression `raw_sql(p.name)` cannot be translated into SQL "
                                        "because raw SQL fragment will be different for each row")
    def test_19(self):
        # raw_sql argument cannot depend on iterator variables
        select(p for p in Person if raw_sql(p.name))[:]

    @db_session
    @raises_exception(ExprEvalError,
                      "`raw_sql('p.dob < $x')` raises NameError: global name 'x' is not defined" if PYPY2 else
                      "`raw_sql('p.dob < $x')` raises NameError: name 'x' is not defined")
    def test_20(self):
        # testing for situation where parameter variable is missing
        select(p for p in Person if raw_sql('p.dob < $x'))[:]

    @db_session
    def test_21(self):
        x = None
        persons = select(p for p in Person if p.id == 1 and raw_sql('$x') is None)[:]
        self.assertEqual(persons, [Person[1]])


if __name__ == '__main__':
    unittest.main()



from pony.orm import *

db = Database(
    provider='cockroach', user='root', host='localhost', port=26257, sslmode='disable', database='test'
)

class T1(db.Entity):
    a = PrimaryKey(int)
    b = Required(str)

db.generate_mapping(check_tables=True)

sql_debug(True)

@db_session(retry=10)
def f():
    x = T1[10]
    x.b += '#'
    flush()
    import time
    for i in range(30):
        print('((%d))' % i)
        time.sleep(1)
    y = T1[20]
    y.b += '@'
    flush()

if __name__ == '__main__':
    f()



from __future__ import absolute_import, print_function

from decimal import Decimal
from datetime import datetime

from pony.converting import str2datetime
from pony.orm import *

db = Database("sqlite", "estore.sqlite", create_db=True)

class Customer(db.Entity):
    email = Required(str, unique=True)
    password = Required(str)
    name = Required(str)
    country = Required(str)
    address = Required(str)
    cart_items = Set("CartItem")
    orders = Set("Order")

class Product(db.Entity):
    id = PrimaryKey(int, auto=True)
    name = Required(str)
    categories = Set("Category")
    description = Optional(str)
    picture = Optional(buffer)
    price = Required(Decimal)
    quantity = Required(int)
    cart_items = Set("CartItem")
    order_items = Set("OrderItem")

class CartItem(db.Entity):
    quantity = Required(int)
    customer = Required(Customer)
    product = Required(Product)

class OrderItem(db.Entity):
    quantity = Required(int)
    price = Required(Decimal)
    order = Required("Order")
    product = Required(Product)
    PrimaryKey(order, product)

class Order(db.Entity):
    id = PrimaryKey(int, auto=True)
    state = Required(str)
    date_created = Required(datetime)
    date_shipped = Optional(datetime)
    date_delivered = Optional(datetime)
    total_price = Required(Decimal)
    customer = Required(Customer)
    items = Set(OrderItem)

class Category(db.Entity):
    name = Required(str, unique=True)
    products = Set(Product)

sql_debug(True)

db.generate_mapping(create_tables=True)

# Order states
CREATED = 'CREATED'
SHIPPED = 'SHIPPED'
DELIVERED = 'DELIVERED'
CANCELLED = 'CANCELLED'

@db_session
def populate_database():
    c1 = Customer(email='john@example.com', password='***',
                  name='John Smith', country='USA', address='address 1')

    c2 = Customer(email='matthew@example.com', password='***',
                  name='Matthew Reed', country='USA', address='address 2')

    c3 = Customer(email='chuanqin@example.com', password='***',
                  name='Chuan Qin', country='China', address='address 3')

    c4 = Customer(email='rebecca@example.com', password='***',
                  name='Rebecca Lawson', country='USA', address='address 4')

    c5 = Customer(email='oliver@example.com', password='***',
                  name='Oliver Blakey', country='UK', address='address 5')

    tablets = Category(name='Tablets')
    flash_drives = Category(name='USB Flash Drives')
    ssd = Category(name='Solid State Drives')
    storage = Category(name='Data Storage')

    p1 = Product(name='Kindle Fire HD', price=Decimal('284.00'), quantity=120,
                 description='Amazon tablet for web, movies, music, apps, '
                             'games, reading and more',
                 categories=[tablets])

    p2 = Product(name='Apple iPad with Retina Display MD513LL/A (16GB, Wi-Fi, White)',
                 price=Decimal('478.50'), quantity=180,
                 description='iPad with Retina display now features an A6X chip, '
                             'FaceTime HD camera, and faster Wi-Fi',
                 categories=[tablets])

    p3 = Product(name='SanDisk Cruzer 16 GB USB Flash Drive', price=Decimal('9.99'),
                 quantity=400, description='Take it all with you on reliable '
                                           'SanDisk USB flash drive',
                 categories=[flash_drives, storage])

    p4 = Product(name='Kingston Digital DataTraveler SE9 16GB USB 2.0',
                 price=Decimal('9.98'), quantity=350,
                 description='Convenient - small, capless and pocket-sized '
                             'for easy transportability',
                 categories=[flash_drives, storage])

    p5 = Product(name='Samsung 840 Series 2.5 inch 120GB SATA III SSD',
                 price=Decimal('98.95'), quantity=0,
                 description='Enables you to boot up your computer '
                             'in as little as 15 seconds',
                 categories=[ssd, storage])

    p6 = Product(name='Crucial m4 256GB 2.5-Inch SSD SATA 6Gb/s CT256M4SSD2',
                 price=Decimal('188.67'), quantity=60,
                 description='The award-winning SSD delivers '
                             'powerful performance gains for SATA 6Gb/s systems',
                 categories=[ssd, storage])

    CartItem(customer=c1, product=p1, quantity=1)
    CartItem(customer=c1, product=p2, quantity=1)
    CartItem(customer=c2, product=p5, quantity=2)

    o1 = Order(customer=c1, total_price=Decimal('292.00'), state=DELIVERED,
               date_created=str2datetime('2012-10-20 15:22:00'),
               date_shipped=str2datetime('2012-10-21 11:34:00'),
               date_delivered=str2datetime('2012-10-26 17:23:00'))

    OrderItem(order=o1, product=p1, price=Decimal('274.00'), quantity=1)
    OrderItem(order=o1, product=p4, price=Decimal('9.98'), quantity=2)

    o2 = Order(customer=c1, total_price=Decimal('478.50'), state=DELIVERED,
               date_created=str2datetime('2013-01-10 09:40:00'),
               date_shipped=str2datetime('2013-01-10 14:03:00'),
               date_delivered=str2datetime('2013-01-13 11:57:00'))

    OrderItem(order=o2, product=p2, price=Decimal('478.50'), quantity=1)

    o3 = Order(customer=c2, total_price=Decimal('680.50'), state=DELIVERED,
               date_created=str2datetime('2012-11-03 12:10:00'),
               date_shipped=str2datetime('2012-11-04 11:47:00'),
               date_delivered=str2datetime('2012-11-07 18:55:00'))

    OrderItem(order=o3, product=p2, price=Decimal('478.50'), quantity=1)
    OrderItem(order=o3, product=p4, price=Decimal('9.98'), quantity=2)
    OrderItem(order=o3, product=p6, price=Decimal('199.00'), quantity=1)

    o4 = Order(customer=c3, total_price=Decimal('99.80'), state=SHIPPED,
               date_created=str2datetime('2013-03-11 19:33:00'),
               date_shipped=str2datetime('2013-03-12 09:40:00'))

    OrderItem(order=o4, product=p4, price=Decimal('9.98'), quantity=10)

    o5 = Order(customer=c4, total_price=Decimal('722.00'), state=CREATED,
               date_created=str2datetime('2013-03-15 23:15:00'))

    OrderItem(order=o5, product=p1, price=Decimal('284.00'), quantity=1)
    OrderItem(order=o5, product=p2, price=Decimal('478.50'), quantity=1)

@db_session
def test_queries():

    print('All USA customers')
    print()
    result = select(c for c in Customer if c.country == 'USA')[:]

    print(result)
    print()

    print('The number of customers for each country')
    print()
    result = select((c.country, count(c)) for c in Customer)[:]

    print(result)
    print()

    print('Max product price')
    print()
    result = max(p.price for p in Product)

    print(result)
    print()

    print('Max SSD price')
    print()
    result = max(p.price for p in Product for cat in p.categories if cat.name == 'Solid State Drives')

    print(result)
    print()

    print('Three most expensive products:')
    print()
    result = select(p for p in Product).order_by(desc(Product.price))[:3]

    print(result)
    print()

    print('Out of stock products')
    print()
    result = select(p for p in Product if p.quantity == 0)[:]

    print(result)
    print()

    print('Most popular product')
    print()
    result = select(p for p in Product).order_by(lambda p: desc(sum(p.order_items.quantity))).first()

    print(result)
    print()

    print('Products that have never been ordered')
    print()
    result = select(p for p in Product if not p.order_items)[:]

    print(result)
    print()

    print('Customers who made several orders')
    print()
    result = select(c for c in Customer if count(c.orders) > 1)[:]

    print(result)
    print()

    print('Three most valuable customers')
    print()
    result = select(c for c in Customer).order_by(lambda c: desc(sum(c.orders.total_price)))[:3]

    print(result)
    print()

    print('Customers whose orders were shipped')
    print()
    result = select(c for c in Customer if SHIPPED in c.orders.state)[:]

    print(result)
    print()

    print('The same query with the INNER JOIN instead of IN')
    print()
    result = select(c for c in Customer if JOIN(SHIPPED in c.orders.state))[:]

    print(result)
    print()

    print('Customers with no orders')
    print()
    result = select(c for c in Customer if not c.orders)[:]

    print(result)
    print()

    print('The same query with the LEFT JOIN instead of NOT EXISTS')
    print()
    result = left_join(c for c in Customer for o in c.orders if o is None)[:]

    print(result)
    print()

    print('Customers which ordered several different tablets')
    print()
    result = select(c for c in Customer
                      for p in c.orders.items.product
                      if 'Tablets' in p.categories.name and count(p) > 1)[:]

    print(result)
    print()

    print('Customers which ordered several products from the same category')
    print()
    result = select((customer, category.name)
                    for customer in Customer
                    for product in customer.orders.items.product
                    for category in product.categories
                    if count(product) > 1)[:]

    print(result)
    print()

    print('Customers which ordered several products from the same category in the same order')
    print()
    result = select((customer, order, category.name)
                    for customer in Customer
                    for order in customer.orders
                    for product in order.items.product
                    for category in product.categories
                    if count(product) > 1)[:]

    print(result)
    print()

    print('Products whose price varies over time')
    print()
    result = select(p.name for p in Product if count(p.order_items.price) > 1)[:]

    print(result)
    print()

    print('The same query, but with min and max price for each product')
    print()
    result = select((p.name, min(p.order_items.price), max(p.order_items.price))
                    for p in Product if count(p.order_items.price) > 1)[:]

    print(result)
    print()

    print('Orders with a discount (order total price < sum of order item prices)')
    print()
    result = select(o for o in Order if o.total_price < sum(o.items.price * o.items.quantity))[:]

    print(result)
    print()


if __name__ == '__main__':
    with db_session:
        if Customer.select().first() is None:
            populate_database()
    test_queries()



import unittest

from pony import orm
from pony.orm.tests import setup_database, teardown_database

db = orm.Database()


class Foo(db.Entity):
    id = orm.PrimaryKey(int)
    b = orm.Optional(orm.buffer)


class Bar(db.Entity):
    b = orm.PrimaryKey(orm.buffer)


class Baz(db.Entity):
    id = orm.PrimaryKey(int)
    b = orm.Optional(orm.buffer, unique=True)


buf = orm.buffer(b'123')

class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with orm.db_session:
            Foo(id=1, b=buf)
            Bar(b=buf)
            Baz(id=1, b=buf)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def test_1(self):  # Bug #355
        with orm.db_session:
            Bar[buf]

    def test_2(self):  # Regression after #355 fix
        with orm.db_session:
            result = orm.select(bar.b for bar in Foo)[:]
            self.assertEqual(result, [buf])

    def test_3(self):  # Bug #390
        with orm.db_session:
            Baz.get(b=buf)



from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
from pony.orm.tests.testutils import raises_exception
from pony.orm.tests import setup_database, teardown_database, only_for

db = Database()


class Person(db.Entity):
    name = Required(unicode)
    spouse = Optional('Person', reverse='spouse')


class TestSymmetricOne2One(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        with db_session:
            db.execute('update person set spouse=null')
            db.execute('delete from person')
            db.insert(Person, id=1, name='A')
            db.insert(Person, id=2, name='B', spouse=1)
            db.execute('update person set spouse=2 where id=1')
            db.insert(Person, id=3, name='C')
            db.insert(Person, id=4, name='D', spouse=3)
            db.execute('update person set spouse=4 where id=3')
            db.insert(Person, id=5, name='E', spouse=None)
        db_session.__enter__()
    def tearDown(self):
        db_session.__exit__()
    def test1(self):
        p1 = Person[1]
        p2 = Person[2]
        p5 = Person[5]
        p1.spouse = p5
        commit()
        self.assertEqual(p1._vals_.get(Person.spouse), p5)
        self.assertEqual(p5._vals_.get(Person.spouse), p1)
        self.assertEqual(p2._vals_.get(Person.spouse), None)
        data = db.select('spouse from person order by id')
        self.assertEqual([5, None, 4, 3, 1], data)
    def test2(self):
        p1 = Person[1]
        p2 = Person[2]
        p1.spouse = None
        commit()
        self.assertEqual(p1._vals_.get(Person.spouse), None)
        self.assertEqual(p2._vals_.get(Person.spouse), None)
        data = db.select('spouse from person order by id')
        self.assertEqual([None, None, 4, 3, None], data)
    def test3(self):
        p1 = Person[1]
        p2 = Person[2]
        p3 = Person[3]
        p4 = Person[4]
        p1.spouse = p3
        commit()
        self.assertEqual(p1._vals_.get(Person.spouse), p3)
        self.assertEqual(p2._vals_.get(Person.spouse), None)
        self.assertEqual(p3._vals_.get(Person.spouse), p1)
        self.assertEqual(p4._vals_.get(Person.spouse), None)
        data = db.select('spouse from person order by id')
        self.assertEqual([3, None, 1, None, None], data)
    def test4(self):
        persons = set(select(p for p in Person if p.spouse.name in ('B', 'D')))
        self.assertEqual(persons, {Person[1], Person[3]})
    @raises_exception(UnrepeatableReadError, 'Multiple Person objects linked with the same Person[2] object. '
                                             'Maybe Person.spouse attribute should be Set instead of Optional')
    def test5(self):
        db.execute('update person set spouse = 3 where id = 2')
        p1 = Person[1]
        p1.spouse
        p2 = Person[2]
        p2.name
    def test6(self):
        db.execute('update person set spouse = 3 where id = 2')
        p1 = Person[1]
        p2 = Person[2]
        p2.name
        p1.spouse
        self.assertEqual(p2._vals_.get(Person.spouse), p1)

if __name__ == '__main__':
    unittest.main()



import unittest

from pony.orm import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Country(db.Entity):
    id = PrimaryKey(int)
    name = Required(str)
    persons = Set("Person")

class Person(db.Entity):
    id = PrimaryKey(int)
    name = Required(str)
    country = Required(Country)

class TestProxy(unittest.TestCase):
    def setUp(self):
        setup_database(db)
        with db_session:
            c1 = Country(id=1, name='Russia')
            c2 = Country(id=2, name='Japan')
            Person(id=1, name='Alexander Nevskiy', country=c1)
            Person(id=2, name='Raikou Minamoto', country=c2)
            Person(id=3, name='Ibaraki Douji', country=c2)

    def tearDown(self):
        teardown_database(db)

    def test_1(self):
        with db_session:
            p = make_proxy(Person[2])

        with db_session:
            x1 = db.local_stats[None].db_count  # number of queries
            # it is possible to access p attributes in a new db_session
            name = p.name
            country = p.country
            x2 = db.local_stats[None].db_count

        # p.name and p.country are loaded with a single query
        self.assertEqual(x1, x2-1)

    def test_2(self):
        with db_session:
            p = make_proxy(Person[2])
            name = p.name
            country = p.country

        with db_session:
            x1 = db.local_stats[None].db_count
            name = p.name
            country = p.country
            x2 = db.local_stats[None].db_count

        # attribute values from the first db_session should be ignored and loaded again
        self.assertEqual(x1, x2-1)

    def test_3(self):
        with db_session:
            p = Person[2]
            proxy = make_proxy(p)

        with db_session:
            p2 = Person[2]
            name1 = 'Tamamo no Mae'
            # It is possible to assign new attribute values to a proxy object
            p2.name = name1
            name2 = proxy.name

        self.assertEqual(name1, name2)


    def test_4(self):
        with db_session:
            p = Person[2]
            proxy = make_proxy(p)

        with db_session:
            p2 = Person[2]
            name1 = 'Tamamo no Mae'
            p2.name = name1

        with db_session:
            # new attribute value was successfully stored in the database
            name2 = proxy.name

        self.assertEqual(name1, name2)

    def test_5(self):
        with db_session:
            p = Person[2]
            r = repr(p)
            self.assertEqual(r, 'Person[2]')

            proxy = make_proxy(p)
            r = repr(proxy)
            # proxy object has specific repr
            self.assertEqual(r, '<EntityProxy(Person[2])>')

        r = repr(proxy)
        # repr of proxy object can be used outside of db_session
        self.assertEqual(r, '<EntityProxy(Person[2])>')

        del p
        r = repr(proxy)
        # repr works even if the original object was deleted
        self.assertEqual(r, '<EntityProxy(Person[2])>')


    def test_6(self):
        with db_session:
            p = Person[2]
            proxy = make_proxy(p)
            proxy.name = 'Okita Souji'
            # after assignment, the attribute value is the same for the proxy and for the original object
            self.assertEqual(proxy.name, 'Okita Souji')
            self.assertEqual(p.name, 'Okita Souji')


    def test_7(self):
        with db_session:
            p = Person[2]
            proxy = make_proxy(p)
            proxy.name = 'Okita Souji'
            # after assignment, the attribute value is the same for the proxy and for the original object
            self.assertEqual(proxy.name, 'Okita Souji')
            self.assertEqual(p.name, 'Okita Souji')


    def test_8(self):
        with db_session:
            c1 = Country[1]
            c1_proxy = make_proxy(c1)
            p2 = Person[2]
            self.assertNotEqual(p2.country, c1)
            self.assertNotEqual(p2.country, c1_proxy)
            # proxy can be used in attribute assignment
            p2.country = c1_proxy
            self.assertEqual(p2.country, c1_proxy)
            self.assertIs(p2.country, c1)


    def test_9(self):
        with db_session:
            c2 = Country[2]
            c2_proxy = make_proxy(c2)
            persons = select(p for p in Person if p.country == c2_proxy)
            self.assertEqual({p.id for p in persons}, {2, 3})

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
import pony.orm.decompiling
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database

db = Database()

class Person(db.Entity):
    name = Required(unicode)
    age = Required(int)


class TestFrames(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            p1 = Person(id=1, name='John', age=22)
            p2 = Person(id=2, name='Mary', age=18)
            p3 = Person(id=3, name='Mike', age=25)

    @classmethod
    def tearDownClass(cls):
        db.drop_all_tables(with_all_data=True)

    @db_session
    def test_select(self):
        x = 20
        result = select(p.id for p in Person if p.age > x)[:]
        self.assertEqual(set(result), {1, 3})

    @db_session
    def test_select_str(self):
        x = 20
        result = select('p.id for p in Person if p.age > x')[:]
        self.assertEqual(set(result), {1, 3})

    @db_session
    def test_left_join(self):
        x = 20
        result = left_join(p.id for p in Person if p.age > x)[:]
        self.assertEqual(set(result), {1, 3})

    @db_session
    def test_left_join_str(self):
        x = 20
        result = left_join('p.id for p in Person if p.age > x')[:]
        self.assertEqual(set(result), {1, 3})

    @db_session
    def test_get(self):
        x = 23
        result = get(p.id for p in Person if p.age > x)
        self.assertEqual(result, 3)

    @db_session
    def test_get_str(self):
        x = 23
        result = get('p.id for p in Person if p.age > x')
        self.assertEqual(result, 3)

    @db_session
    def test_exists(self):
        x = 23
        result = exists(p for p in Person if p.age > x)
        self.assertEqual(result, True)

    @db_session
    def test_exists_str(self):
        x = 23
        result = exists('p for p in Person if p.age > x')
        self.assertEqual(result, True)

    @db_session
    def test_entity_get(self):
        x = 23
        p = Person.get(lambda p: p.age > x)
        self.assertEqual(p, Person[3])

    @db_session
    def test_entity_get_str(self):
        x = 23
        p = Person.get('lambda p: p.age > x')
        self.assertEqual(p, Person[3])

    @db_session
    def test_entity_get_by_sql(self):
        x = 25
        p = Person.get_by_sql('select * from Person where age = $x')
        self.assertEqual(p, Person[3])

    @db_session
    def test_entity_select_by_sql(self):
        x = 25
        p = Person.select_by_sql('select * from Person where age = $x')
        self.assertEqual(p, [ Person[3] ])

    @db_session
    def test_entity_exists(self):
        x = 23
        result = Person.exists(lambda p: p.age > x)
        self.assertTrue(result)

    @db_session
    def test_entity_exists_str(self):
        x = 23
        result = Person.exists('lambda p: p.age > x')
        self.assertTrue(result)

    @db_session
    def test_entity_select(self):
        x = 20
        result = Person.select(lambda p: p.age > x)[:]
        self.assertEqual(set(result), {Person[1], Person[3]})

    @db_session
    def test_entity_select_str(self):
        x = 20
        result = Person.select('lambda p: p.age > x')[:]
        self.assertEqual(set(result), {Person[1], Person[3]})

    @db_session
    def test_order_by(self):
        x = 20
        y = -1
        result = Person.select(lambda p: p.age > x).order_by(lambda p: p.age * y)[:]
        self.assertEqual(result, [Person[3], Person[1]])

    @db_session
    def test_order_by_str(self):
        x = 20
        y = -1
        result = Person.select('lambda p: p.age > x').order_by('p.age * y')[:]
        self.assertEqual(result, [Person[3], Person[1]])

    @db_session
    def test_filter(self):
        x = 20
        y = 'M'
        result = Person.select(lambda p: p.age > x).filter(lambda p: p.name.startswith(y))[:]
        self.assertEqual(result, [Person[3]])

    @db_session
    def test_filter_str(self):
        x = 20
        y = 'M'
        result = Person.select('lambda p: p.age > x').filter('p.name.startswith(y)')[:]
        self.assertEqual(result, [Person[3]])

    @db_session
    def test_db_select(self):
        x = 20
        result = db.select('name from Person where age > $x order by name')
        self.assertEqual(result, ['John', 'Mike'])

    @db_session
    def test_db_get(self):
        x = 18
        result = db.get('name from Person where age = $x')
        self.assertEqual(result, 'Mary')

    @db_session
    def test_db_execute(self):
        x = 18
        result = db.execute('select name from Person where age = $x').fetchone()
        self.assertEqual(result, ('Mary',))

    @db_session
    def test_db_exists(self):
        x = 18
        result = db.exists('name from Person where age = $x')
        self.assertEqual(result, True)

    @raises_exception(pony.orm.decompiling.InvalidQuery,
                      'Use generator expression (... for ... in ...) '
                      'instead of list comprehension [... for ... in ...] inside query')
    @db_session
    def test_inner_list_comprehension(self):
        result = select(p.id for p in Person if p.age not in [
            p2.age for p2 in Person if p2.name.startswith('M')])[:]

    @db_session
    def test_outer_list_comprehension(self):
        names = ['John', 'Mary', 'Mike']
        persons = [ Person.select(lambda p: p.name == name).first() for name in names ]
        self.assertEqual(set(p.name for p in persons), {'John', 'Mary', 'Mike'})

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from pony.py23compat import PY2, basestring, unicode, buffer, int_types

from decimal import Decimal
from datetime import datetime, date, time, timedelta
from uuid import UUID

try:
    import psycopg2
except ImportError:
    try:
        from psycopg2cffi import compat
    except ImportError:
        raise ImportError('In order to use PonyORM with CockroachDB please install psycopg2 or psycopg2cffi')
    else:
        compat.register()

from pony.orm.dbproviders.postgres import (
    PGSQLBuilder, PGColumn, PGSchema, PGTranslator, PGProvider,
    PGStrConverter, PGIntConverter, PGRealConverter,
    PGDatetimeConverter, PGTimedeltaConverter,
    PGBlobConverter, PGJsonConverter, PGArrayConverter,
)

from pony.orm import core, dbapiprovider, ormtypes
from pony.orm.core import log_orm
from pony.orm.dbapiprovider import wrap_dbapi_exceptions

NoneType = type(None)

class CRColumn(PGColumn):
    auto_template = 'SERIAL PRIMARY KEY'

class CRSchema(PGSchema):
    column_class = CRColumn

class CRTranslator(PGTranslator):
    pass

class CRSQLBuilder(PGSQLBuilder):
    pass

class CRIntConverter(PGIntConverter):
    signed_types = {None: 'INT', 8: 'INT2', 16: 'INT2', 24: 'INT8', 32: 'INT8', 64: 'INT8'}
    unsigned_types = {None: 'INT', 8: 'INT2', 16: 'INT4', 24: 'INT8', 32: 'INT8'}
    # signed_types = {None: 'INT', 8: 'INT2', 16: 'INT2', 24: 'INT4', 32: 'INT4', 64: 'INT8'}
    # unsigned_types = {None: 'INT', 8: 'INT2', 16: 'INT4', 24: 'INT4', 32: 'INT8'}

class CRBlobConverter(PGBlobConverter):
    def sql_type(converter):
        return 'BYTES'

class CRTimedeltaConverter(PGTimedeltaConverter):
    sql_type_name = 'INTERVAL'

class PGUuidConverter(dbapiprovider.UuidConverter):
    def py2sql(converter, val):
        return val

class CRArrayConverter(PGArrayConverter):
    array_types = {
        int: ('INT', PGIntConverter),
        unicode: ('STRING', PGStrConverter),
        float: ('DOUBLE PRECISION', PGRealConverter)
    }

class CRProvider(PGProvider):
    dbapi_module = psycopg2
    dbschema_cls = CRSchema
    translator_cls = CRTranslator
    sqlbuilder_cls = CRSQLBuilder
    array_converter_cls = CRArrayConverter

    default_schema_name = 'public'

    fk_types = { 'SERIAL' : 'INT8' }

    def normalize_name(provider, name):
        return name[:provider.max_name_len].lower()

    @wrap_dbapi_exceptions
    def set_transaction_mode(provider, connection, cache):
        assert not cache.in_transaction
        db_session = cache.db_session
        if db_session is not None and db_session.ddl:
            cache.immediate = False
        if cache.immediate and connection.autocommit:
            connection.autocommit = False
            if core.local.debug: log_orm('SWITCH FROM AUTOCOMMIT TO TRANSACTION MODE')
        elif not cache.immediate and not connection.autocommit:
            connection.autocommit = True
            if core.local.debug: log_orm('SWITCH TO AUTOCOMMIT MODE')
        if db_session is not None and (db_session.serializable or db_session.ddl):
            cache.in_transaction = True

    converter_classes = [
        (NoneType, dbapiprovider.NoneConverter),
        (bool, dbapiprovider.BoolConverter),
        (basestring, PGStrConverter),
        (int_types, CRIntConverter),
        (float, PGRealConverter),
        (Decimal, dbapiprovider.DecimalConverter),
        (datetime, PGDatetimeConverter),
        (date, dbapiprovider.DateConverter),
        (time, dbapiprovider.TimeConverter),
        (timedelta, CRTimedeltaConverter),
        (UUID, PGUuidConverter),
        (buffer, CRBlobConverter),
        (ormtypes.Json, PGJsonConverter),
    ]

provider_cls = CRProvider



import unittest
from datetime import timedelta

from pony import converting

class Test(unittest.TestCase):
    def test_timestamps_1(self):
        s = '1:2:3'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=1, minutes=2, seconds=3))
        s = '01:02:03'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=1, minutes=2, seconds=3))
        s = '1:2:3.456'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=1, minutes=2, seconds=3, milliseconds=456))
        s = '1:2:3.45678'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=1, minutes=2, seconds=3, microseconds=456780))
        s = '12:34:56'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=12, minutes=34, seconds=56))
        s = '12:34:56.789'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=12, minutes=34, seconds=56, milliseconds=789))
        s = '12:34:56.789123'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(hours=12, minutes=34, seconds=56, microseconds=789123))

    def test_timestamps_2(self):
        s = '-1:2:3'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=1, minutes=2, seconds=3))
        s = '-01:02:03'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=1, minutes=2, seconds=3))
        s = '-1:2:3.456'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=1, minutes=2, seconds=3, milliseconds=456))
        s = '-1:2:3.45678'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=1, minutes=2, seconds=3, microseconds=456780))
        s = '-12:34:56'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=12, minutes=34, seconds=56))
        s = '-12:34:56.789'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=12, minutes=34, seconds=56, milliseconds=789))
        s = '-12:34:56.789123'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(hours=12, minutes=34, seconds=56, microseconds=789123))

    def test_timestamps_3(self):
        s = '0:1:2'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(minutes=1, seconds=2))
        s = '0:1:2.3456'
        td = converting.str2timedelta(s)
        self.assertEqual(td, timedelta(minutes=1, seconds=2, microseconds=345600))

    def test_timestamps_4(self):
        s = '-0:1:2'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(minutes=1, seconds=2))
        s = '-0:1:2.3456'
        td = converting.str2timedelta(s)
        self.assertEqual(td, -timedelta(minutes=1, seconds=2, microseconds=345600))



from __future__ import absolute_import, print_function, division

import unittest
from pony.orm.sqlsymbols import *
from pony.orm.sqlbuilding import SQLBuilder
from pony.orm.dbapiprovider import DBAPIProvider
from pony.orm.tests.testutils import TestPool


class TestFormatStyles(unittest.TestCase):
    def setUp(self):
        self.key1 = 'KEY1'
        self.key2 = 'KEY2'
        self.provider = DBAPIProvider(pony_pool_mockup=TestPool(None))
        self.ast = [ SELECT, [ ALL, [COLUMN, None, 'A']], [ FROM, [None, TABLE, 'T1']],
                     [ WHERE, [ EQ, [COLUMN, None, 'B'], [ PARAM, self.key1 ] ],
                              [ EQ, [COLUMN, None, 'C'], [ PARAM, self.key2 ] ],
                              [ EQ, [COLUMN, None, 'D'], [ PARAM, self.key2 ] ],
                              [ EQ, [COLUMN, None, 'E'], [ PARAM, self.key1 ] ]
                     ]
                   ]
    def test_qmark(self):
        self.provider.paramstyle = 'qmark'
        b = SQLBuilder(self.provider, self.ast)
        self.assertEqual(b.sql, 'SELECT "A"\n'
                                'FROM "T1"\n'
                                'WHERE "B" = ?\n  AND "C" = ?\n  AND "D" = ?\n  AND "E" = ?')
        self.assertEqual(b.layout, [self.key1, self.key2, self.key2, self.key1])
    def test_numeric(self):
        self.provider.paramstyle = 'numeric'
        b = SQLBuilder(self.provider, self.ast)
        self.assertEqual(b.sql, 'SELECT "A"\n'
                                'FROM "T1"\n'
                                'WHERE "B" = :1\n  AND "C" = :2\n  AND "D" = :2\n  AND "E" = :1')
        self.assertEqual(b.layout, [self.key1, self.key2, self.key2, self.key1])
    def test_named(self):
        self.provider.paramstyle = 'named'
        b = SQLBuilder(self.provider, self.ast)
        self.assertEqual(b.sql, 'SELECT "A"\n'
                                'FROM "T1"\n'
                                'WHERE "B" = :p1\n  AND "C" = :p2\n  AND "D" = :p2\n  AND "E" = :p1')
        self.assertEqual(b.layout, [self.key1, self.key2, self.key2, self.key1])
    def test_format(self):
        self.provider.paramstyle = 'format'
        b = SQLBuilder(self.provider, self.ast)
        self.assertEqual(b.sql, 'SELECT "A"\n'
                                'FROM "T1"\n'
                                'WHERE "B" = %s\n  AND "C" = %s\n  AND "D" = %s\n  AND "E" = %s')
        self.assertEqual(b.layout, [self.key1, self.key2, self.key2, self.key1])
    def test_pyformat(self):
        self.provider.paramstyle = 'pyformat'
        b = SQLBuilder(self.provider, self.ast)
        self.assertEqual(b.sql, 'SELECT "A"\n'
                                'FROM "T1"\n'
                                'WHERE "B" = %(p1)s\n  AND "C" = %(p2)s\n  AND "D" = %(p2)s\n  AND "E" = %(p1)s')
        self.assertEqual(b.layout, [self.key1, self.key2, self.key2, self.key1])


if __name__ == "__main__":
    unittest.main()



from pony.py23compat import StringIO
from pony.orm import *
from pony.orm.tests import setup_database, teardown_database

import unittest

db = Database()

class A(db.Entity):
    id = PrimaryKey(int)
    x = Required(bool)
    y = Required(float)


class TestDeduplication(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            a1 = A(id=1, x=False, y=3.0)
            a2 = A(id=2, x=True, y=4.0)
            a3 = A(id=3, x=False, y=1.0)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    @db_session
    def test_1(self):
        a2 = A.get(id=2)
        a1 = A.get(id=1)
        self.assertIs(a1.id, 1)

    @db_session
    def test_2(self):
        a3 = A.get(id=3)
        a1 = A.get(id=1)
        self.assertIs(a1.id, 1)

    @db_session
    def test_3(self):
        q = A.select().order_by(-1)
        stream = StringIO()
        q.show(stream=stream)
        s = stream.getvalue()
        self.assertEqual(s, 'id|x    |y  \n'  
                            '--+-----+---\n'
                            '3 |False|1.0\n'
                            '2 |True |4.0\n'
                            '1 |False|3.0\n')






import unittest

from pony.orm.tests import setup_database, teardown_database
from pony.orm import *

db = Database()


class Person(db.Entity):
    name = Required(str)
    group = Optional(lambda: Group)


class Group(db.Entity):
    title = PrimaryKey(str)
    persons = Set(Person)

    def __len__(self):
        return len(self.persons)


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def test_1(self):
        with db_session:
            p1 = Person(name="Alex")
            p2 = Person(name="Brad")
            p3 = Person(name="Chad")
            p4 = Person(name="Dylan")
            p5 = Person(name="Ethan")

            g1 = Group(title="Foxes")
            g2 = Group(title="Gorillas")

            g1.persons.add(p1)
            g1.persons.add(p2)
            g1.persons.add(p3)
            g2.persons.add(p4)
            g2.persons.add(p5)
            commit()

            foxes = Group['Foxes']
            gorillas = Group['Gorillas']

            self.assertEqual(len(foxes), 3)
            self.assertEqual(len(gorillas), 2)



from __future__ import absolute_import, print_function

from pony.orm.core import *

db = Database()

class Numbers(db.Entity):
    _table_ = "Numbers"
    id = PrimaryKey(int, auto=True)
    int8 = Required(int, size=8)    # TINYINT
    int16 = Required(int, size=16)  # SMALLINT
    int24 = Required(int, size=24)  # MEDIUMINT
    int32 = Required(int, size=32)  # INTEGER
    int64 = Required(int, size=64)  # BIGINT
    uint8 = Required(int, size=8, unsigned=True)    # TINYINT UNSIGNED
    uint16 = Required(int, size=16, unsigned=True)  # SMALLINT UNSIGNED
    uint24 = Required(int, size=24, unsigned=True)  # MEDIUMINT UNSIGNED
    uint32 = Required(int, size=32, unsigned=True)  # INTEGER UNSIGNED
    # uint64 = Required(int, size=64, unsigned=True)  # BIGINT UNSIGNED, supported by MySQL and Oracle

sql_debug(True)  # Output all SQL queries to stdout

db.bind('sqlite', 'test_numbers.sqlite', create_db=True)
#db.bind('mysql', host="localhost", user="pony", passwd="pony", db="test_numbers")
#db.bind('postgres', user='pony', password='pony', host='localhost', database='test_numbers')
#db.bind('oracle', 'test_numbers/pony@localhost')

db.drop_table("Numbers", if_exists=True, with_all_data=True)
db.generate_mapping(create_tables=True)

@db_session
def populate_database():
    lo = Numbers(int8=-128,
                 int16=-32768,
                 int24=-8388608,
                 int32=-2147483648,
                 int64=-9223372036854775808,
                 uint8=0, uint16=0, uint24=0, uint32=0) #, uint64=0)
    hi = Numbers(int8=127,
                 int16=32767,
                 int24=8388607,
                 int32=2147483647,
                 int64=9223372036854775807,
                 uint8=255,
                 uint16=65535,
                 uint24=16777215,
                 uint32=4294967295)
                 # uint64=18446744073709551615)
    commit()

@db_session
def test_data():
    for n in Numbers.select():
        print(n.id, n.int8, n.int16, n.int24, n.int32, n.int64,
              n.uint8, n.uint16, n.uint24, n.uint32) #, n.uint64)

if __name__ == '__main__':
    populate_database()
    test_data()



import unittest

from pony import orm
from pony.orm.tests import setup_database, teardown_database

db = orm.Database()


class Person(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    name = orm.Required(str)
    orm.composite_key(id, name)


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def test_1(self):
        table = db.schema.tables[Person._table_]
        pk_column = table.column_dict[Person.id.column]
        self.assertTrue(pk_column.is_pk)

        with orm.db_session:
            p1 = Person(name='John')
            p2 = Person(name='Mike')



from __future__ import absolute_import, print_function, division

import unittest

from pony.orm.core import *
from pony.orm.tests.testutils import *
from pony.orm.tests import setup_database, teardown_database, only_for

db = Database()

class Student(db.Entity):
    name = Required(str)
    unstripped = Required(str, autostrip=False)
    foo = Optional(str)
    bar = Optional(str)

class TestStringMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            Student(id=1, name="Ann", unstripped="Ann", foo='Abcdef', bar='b%d')
            Student(id=2, name="Bob", unstripped=" Bob ", foo='Ab%def', bar='b%d')
            Student(id=3, name="Beth", unstripped="  Beth  ", foo='Ab_def', bar='b%d')
            Student(id=4, name="Jonathan", unstripped="\nJonathan\n")
            Student(id=5, name="Pete", unstripped="\n Pete\n ")

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_getitem_1(self):
        result = set(select(s for s in Student if s.name[1] == 'o'))
        self.assertEqual(result, {Student[2], Student[4]})
        x = 1
        result = set(select(s for s in Student if s.name[x] == 'o'))
        self.assertEqual(result, {Student[2], Student[4]})

    def test_getitem_2(self):
        result = set(select(s for s in Student if s.name[-1] == 'n'))
        self.assertEqual(result, {Student[1], Student[4]})
        x = -1
        result = set(select(s for s in Student if s.name[x] == 'n'))
        self.assertEqual(result, {Student[1], Student[4]})

    def test_getitem_3(self):
        result = set(select(s for s in Student if s.name[-2] == 't'))
        self.assertEqual(result, {Student[3], Student[5]})
        x = -2
        result = set(select(s for s in Student if s.name[x] == 't'))
        self.assertEqual(result, {Student[3], Student[5]})

    def test_getitem_4(self):
        result = set(select(s for s in Student if s.name[-s.id] == 'n'))
        self.assertEqual(result, {Student[1]})

    def test_slice_1(self):
        result = set(select(s for s in Student if s.name[:] == "Ann"))
        self.assertEqual(result, {Student[1]})
        result = set(select(s for s in Student if s.name[0:] == "Ann"))
        self.assertEqual(result, {Student[1]})

    def test_slice_2(self):
        result = set(select(s for s in Student if s.name[:3] == "Jon"))
        self.assertEqual(result, {Student[4]})
        result = set(select(s for s in Student if s.name[0:3] == "Jon"))
        self.assertEqual(result, {Student[4]})
        x = 0
        y = 3
        result = set(select(s for s in Student if s.name[:y] == "Jon"))
        self.assertEqual(result, {Student[4]})
        result = set(select(s for s in Student if s.name[x:y] == "Jon"))
        self.assertEqual(result, {Student[4]})
        result = set(select(s for s in Student if s.name[x:3] == "Jon"))
        self.assertEqual(result, {Student[4]})

    def test_slice_3(self):
        result = set(select(s for s in Student if s.name[0:10] == "Ann"))
        self.assertEqual(result, {Student[1]})
        x = 10
        result = set(select(s for s in Student if s.name[0:x] == "Ann"))
        self.assertEqual(result, {Student[1]})
        result = set(select(s for s in Student if s.name[:x] == "Ann"))
        self.assertEqual(result, {Student[1]})

    def test_slice_8(self):
        result = set(select(s for s in Student if s.name[1:] == "nn"))
        self.assertEqual(result, {Student[1]})
        x = 1
        result = set(select(s for s in Student if s.name[x:] == "nn"))
        self.assertEqual(result, {Student[1]})

    def test_slice_10(self):
        result = set(select(s for s in Student if s.name[1:3] == "et"))
        self.assertEqual(result, {Student[3], Student[5]})
        x = 1
        y = 3
        result = set(select(s for s in Student if s.name[x:y] == "et"))
        self.assertEqual(result, {Student[3], Student[5]})

    def test_slice_11(self):
        x = 10
        y = 20
        result = set(select(s for s in Student if s.name[x:y] == ''))
        self.assertEqual(result, {Student[1], Student[2], Student[3], Student[4], Student[5]})

    def test_slice_12(self):
        result = set(select(s for s in Student if s.name[-2:] == "nn"))
        self.assertEqual(result, {Student[1]})

    def test_slice_13(self):
        result = set(select(s for s in Student if s.name[:-1] == "Ann"))
        self.assertEqual(result, {Student[1]})
        result = set(select(s for s in Student if s.name[0:-1] == "Ann"))
        self.assertEqual(result, {Student[1]})
        x = 0
        y = -1
        result = set(select(s for s in Student if s.name[x:y] == "Ann"))
        self.assertEqual(result, {Student[1]})

    def test_slice_14(self):
        result = set(select(s for s in Student if s.name[-4:-2] == "th"))
        self.assertEqual(result, {Student[4]})
        x = -4
        y = -2
        result = set(select(s for s in Student if s.name[x:y] == "th"))
        self.assertEqual(result, {Student[4]})

    def test_slice_15(self):
        result = set(select(s for s in Student if s.name[4:-2] == "th"))
        self.assertEqual(result, {Student[4]})
        x = 4
        y = -2
        result = set(select(s for s in Student if s.name[x:y] == "th"))
        self.assertEqual(result, {Student[4]})

    def test_slice_16(self):
        result = list(select(s.name[-2:3] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'nn', 'ob', 't', 't'])
        x = -2
        y = 3
        result = list(select(s.name[x:y] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'nn', 'ob', 't', 't'])

    def test_slice_17(self):
        result = list(select(s.name[s.id:5] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'b', 'h', 'nn', 't'])
        x = 5
        result = list(select(s.name[s.id:x] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'b', 'h', 'nn', 't'])

    def test_slice_18(self):
        result = list(select(s.name[-s.id:5] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['Pete', 'eth', 'n', 'ob', 't'])
        x = 5
        result = list(select(s.name[-s.id:x] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['Pete', 'eth', 'n', 'ob', 't'])

    def test_slice_19a(self):
        result = list(select(s.name[s.id:] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'b', 'h', 'nn', 'than'])

    def test_slice_19b(self):
        result = list(select(s.name[s.id:-1] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', '', 'n', 'tha'])
        x = -1
        result = list(select(s.name[s.id:x] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', '', 'n', 'tha'])

    def test_slice_19c(self):
        result = list(select(s.name[s.id:-2] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', '', '', 'th'])
        x = -2
        result = list(select(s.name[s.id:x] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', '', '', 'th'])

    def test_slice_20a(self):
        result = list(select(s.name[-s.id:] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['Pete', 'eth', 'n', 'ob', 'than'])

    def test_slice_20b(self):
        result = list(select(s.name[-s.id:-1] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'Pet', 'et', 'o', 'tha'])
        x = -1
        result = list(select(s.name[-s.id:x] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'Pet', 'et', 'o', 'tha'])

    def test_slice_20c(self):
        result = list(select(s.name[-s.id:-2] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', 'Pe', 'e', 'th'])
        x = -2
        result = list(select(s.name[-s.id:x] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', 'Pe', 'e', 'th'])

    def test_slice_21(self):
        result = list(select(s.name[1:s.id] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'et', 'ete', 'o', 'ona'])
        x = 1
        result = list(select(s.name[x:s.id] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'et', 'ete', 'o', 'ona'])

    def test_slice_22(self):
        result = list(select(s.name[-3:s.id] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'A', 'Bo', 'et', 'ete'])
        x = -3
        result = list(select(s.name[x:s.id] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'A', 'Bo', 'et', 'ete'])

    def test_slice_23(self):
        result = list(select(s.name[s.id:s.id+3] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'b', 'h', 'nn', 'tha'])

    def test_slice_24(self):
        result = list(select(s.name[-s.id*2:-s.id] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', 'B', 'B', 'Jona', 'n'])

    def test_slice_25(self):
        result = list(select(s.name[s.id:-s.id+3] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['', '', '', 'n', 'tha'])

    def test_slice_26(self):
        result = list(select(s.name[-s.id:s.id+3] for s in Student).without_distinct())
        self.assertEqual(sorted(result), ['Pete', 'eth', 'n', 'ob', 'tha'])

    def test_nonzero(self):
        result = set(select(s for s in Student if s.foo))
        self.assertEqual(result, {Student[1], Student[2], Student[3]})

    def test_add(self):
        name = 'Bethy'
        result = set(select(s for s in Student if s.name + "y" == name))
        self.assertEqual(result, {Student[3]})

    def test_contains_1(self):
        result = set(select(s for s in Student if 'o' in s.name))
        self.assertEqual(result, {Student[2], Student[4]})

    def test_contains_2(self):
        result = set(select(s for s in Student if 'an' in s.name))
        self.assertEqual(result, {Student[4]})

    def test_contains_3(self):
        x = 'an'
        result = set(select(s for s in Student if x in s.name))
        self.assertEqual(result, {Student[4]})

    def test_contains_4(self):
        x = 'an'
        result = set(select(s for s in Student if x not in s.name))
        self.assertEqual(result, {Student[1], Student[2], Student[3], Student[5]})

    def test_contains_5(self):
        result = set(select(s for s in Student if '%' in s.foo))
        self.assertEqual(result, {Student[2]})

    def test_contains_6(self):
        x = '%'
        result = set(select(s for s in Student if x in s.foo))
        self.assertEqual(result, {Student[2]})

    def test_contains_7(self):
        result = set(select(s for s in Student if '_' in s.foo))
        self.assertEqual(result, {Student[3]})

    def test_contains_8(self):
        x = '_'
        result = set(select(s for s in Student if x in s.foo))
        self.assertEqual(result, {Student[3]})

    def test_contains_9(self):
        result = set(select(s for s in Student if s.foo in 'Abcdef'))
        self.assertEqual(result, {Student[1], Student[4], Student[5]})

    def test_contains_10(self):
        result = set(select(s for s in Student if s.bar in s.foo))
        self.assertEqual(result, {Student[2], Student[4], Student[5]})

    def test_startswith_1(self):
        students = set(select(s for s in Student if s.name.startswith('B')))
        self.assertEqual(students, {Student[2], Student[3]})

    def test_startswith_2(self):
        students = set(select(s for s in Student if not s.name.startswith('B')))
        self.assertEqual(students, {Student[1], Student[4], Student[5]})

    def test_startswith_3(self):
        students = set(select(s for s in Student if not not s.name.startswith('B')))
        self.assertEqual(students, {Student[2], Student[3]})

    def test_startswith_4(self):
        students = set(select(s for s in Student if not not not s.name.startswith('B')))
        self.assertEqual(students, {Student[1], Student[4], Student[5]})

    def test_startswith_5(self):
        x = "Pe"
        students = select(s for s in Student if s.name.startswith(x))[:]
        self.assertEqual(students, [Student[5]])

    def test_endswith_1(self):
        students = set(select(s for s in Student if s.name.endswith('n')))
        self.assertEqual(students, {Student[1], Student[4]})

    def test_endswith_2(self):
        x = "te"
        students = select(s for s in Student if s.name.endswith(x))[:]
        self.assertEqual(students, [Student[5]])

    def test_strip_1(self):
        students = select(s for s in Student if s.name.strip() == 'Beth')[:]
        self.assertEqual(students, [Student[3]])

    def test_rstrip_1(self):
        students = select(s for s in Student if s.name.rstrip('n') == 'A')[:]
        self.assertEqual(students, [Student[1]])

    def test_rstrip_2(self):
        x = 'n'
        students = select(s for s in Student if s.name.rstrip(x) == 'A')[:]
        self.assertEqual(students, [Student[1]])

    def test_lstrip(self):
        students = select(s for s in Student if s.name.lstrip('P') == 'ete')[:]
        self.assertEqual(students, [Student[5]])

    def test_upper(self):
        result = select(s for s in Student if s.name.upper() == "ANN")[:]
        self.assertEqual(result, [Student[1]])

    def test_lower(self):
        result = select(s for s in Student if s.name.lower() == "ann")[:]
        self.assertEqual(result, [Student[1]])

if __name__ == "__main__":
    unittest.main()



import sys, platform

PY2 = sys.version_info[0] == 2
PYPY = platform.python_implementation() == 'PyPy'
PYPY2 = PYPY and PY2
PY37 = sys.version_info[:2] >= (3, 7)

if PY2:
    from future_builtins import zip as izip, map as imap
    import __builtin__ as builtins
    import cPickle as pickle
    from cStringIO import StringIO
    xrange = xrange
    basestring = basestring
    unicode = unicode
    buffer = buffer
    int_types = (int, long)
    cmp = cmp

    def iteritems(dict):
        return dict.iteritems()

    def itervalues(dict):
        return dict.itervalues()

    def items_list(dict):
        return dict.items()

    def values_list(dict):
        return dict.values()

else:
    import builtins, pickle
    from io import StringIO
    izip, imap, xrange = zip, map, range
    basestring = str
    unicode = str
    buffer = bytes
    int_types = (int,)

    def cmp(a, b):
        return (a > b) - (a < b)

    def iteritems(dict):
        return iter(dict.items())

    def itervalues(dict):
        return iter(dict.values())

    def items_list(dict):
        return list(dict.items())

    def values_list(dict):
        return list(dict.values())

# Armin's recipe from http://lucumr.pocoo.org/2013/5/21/porting-to-python-3-redux/
def with_metaclass(meta, *bases):
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__
        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})
